# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""HESS Galactic plane survey (HGPS) catalog."""

import numpy as np
import astropy.units as u
from astropy.coordinates import Angle
from astropy.modeling.models import Gaussian1D
from astropy.table import Table
from gammapy.estimators import FluxPoints
from gammapy.maps import MapAxis, RegionGeom
from gammapy.modeling.models import Model, Models, SkyModel
from gammapy.utils.interpolation import ScaledRegularGridInterpolator
from gammapy.utils.scripts import make_path
from gammapy.utils.table import table_row_to_dict
from .core import SourceCatalog, SourceCatalogObject, format_flux_points_table

__all__ = [
    "SourceCatalogHGPS",
    "SourceCatalogLargeScaleHGPS",
    "SourceCatalogObjectHGPS",
    "SourceCatalogObjectHGPSComponent",
]

# Flux factor, used for printing
FF = 1e-12

# Multiplicative factor to go from cm^-2 s^-1 to % Crab for integral flux > 1 TeV
# Here we use the same Crab reference that's used in the HGPS paper
# CRAB = crab_integral_flux(energy_min=1, reference='hess_ecpl')
FLUX_TO_CRAB = 100 / 2.26e-11
FLUX_TO_CRAB_DIFF = 100 / 3.5060459323111307e-11


class SourceCatalogObjectHGPSComponent(SourceCatalogObject):
    """One Gaussian component from the HGPS catalog.

    See Also
    --------
    SourceCatalogHGPS, SourceCatalogObjectHGPS
    """

    _source_name_key = "Component_ID"

    def __init__(self, data):
        self.data = data

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name!r})"

    def __str__(self):
        """Pretty-print source data."""
        d = self.data
        ss = "Component {}:\n".format(d["Component_ID"])
        fmt = "{:<20s} : {:8.3f} +/- {:.3f} deg\n"
        ss += fmt.format("GLON", d["GLON"].value, d["GLON_Err"].value)
        ss += fmt.format("GLAT", d["GLAT"].value, d["GLAT_Err"].value)
        fmt = "{:<20s} : {:.3f} +/- {:.3f} deg\n"
        ss += fmt.format("Size", d["Size"].value, d["Size_Err"].value)
        val, err = d["Flux_Map"].value, d["Flux_Map_Err"].value
        fmt = "{:<20s} : ({:.2f} +/- {:.2f}) x 10^-12 cm^-2 s^-1 = ({:.1f} +/- {:.1f}) % Crab"
        ss += fmt.format(
            "Flux (>1 TeV)", val / FF, err / FF, val * FLUX_TO_CRAB, err * FLUX_TO_CRAB
        )
        return ss

    @property
    def name(self):
        """Source name as a string."""
        return self.data[self._source_name_key]

    def spatial_model(self):
        """Component spatial model as a `~gammapy.modeling.models.GaussianSpatialModel` object."""
        d = self.data
        tag = "GaussianSpatialModel"
        pars = {
            "lon_0": d["GLON"],
            "lat_0": d["GLAT"],
            "sigma": d["Size"],
            "frame": "galactic",
        }
        errs = {"lon_0": d["GLON_Err"], "lat_0": d["GLAT_Err"], "sigma": d["Size_Err"]}
        model = Model.create(tag, "spatial", **pars)

        for name, value in errs.items():
            model.parameters[name].error = value

        return model


class SourceCatalogObjectHGPS(SourceCatalogObject):
    """One object from the HGPS catalog.

    The catalog is represented by `SourceCatalogHGPS`.
    Examples are given there.
    """

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name!r})"

    def __str__(self):
        return self.info()

    @property
    def flux_points(self):
        """Flux points as a `~gammapy.estimators.FluxPoints` object."""
        reference_model = SkyModel(spectral_model=self.spectral_model(), name=self.name)
        return FluxPoints.from_table(
            self.flux_points_table,
            reference_model=reference_model,
        )

    def info(self, info="all"):
        """Information string.

        Parameters
        ----------
        info : {'all', 'basic', 'map', 'spec', 'flux_points', 'components', 'associations', 'id'}
            Comma separated list of options.
        """
        if info == "all":
            info = "basic,associations,id,map,spec,flux_points,components"

        ss = ""
        ops = info.split(",")
        if "basic" in ops:
            ss += self._info_basic()
        if "map" in ops:
            ss += self._info_map()
        if "spec" in ops:
            ss += self._info_spec()
        if "flux_points" in ops:
            ss += self._info_flux_points()
        if "components" in ops and hasattr(self, "components"):
            ss += self._info_components()
        if "associations" in ops:
            ss += self._info_associations()
        if "id" in ops and hasattr(self, "identification_info"):
            ss += self._info_id()
        return ss

    def _info_basic(self):
        """Print basic information."""
        d = self.data
        ss = "\n*** Basic info ***\n\n"
        ss += "Catalog row index (zero-based) : {}\n".format(self.row_index)
        ss += "{:<20s} : {}\n".format("Source name", self.name)

        ss += "{:<20s} : {}\n".format("Analysis reference", d["Analysis_Reference"])
        ss += "{:<20s} : {}\n".format("Source class", d["Source_Class"])
        ss += "{:<20s} : {}\n".format("Identified object", d["Identified_Object"])
        ss += "{:<20s} : {}\n".format("Gamma-Cat id", d["Gamma_Cat_Source_ID"])
        ss += "\n"
        return ss

    def _info_associations(self):
        ss = "\n*** Source associations info ***\n\n"
        lines = self.associations.pformat(max_width=-1, max_lines=-1)
        ss += "\n".join(lines)
        return ss + "\n"

    def _info_id(self):
        ss = "\n*** Source identification info ***\n\n"
        ss += "\n".join(f"{k}: {v}" for k, v in self.identification_info.items())
        return ss + "\n"

    def _info_map(self):
        """Print information from map analysis."""
        d = self.data
        ss = "\n*** Info from map analysis ***\n\n"

        ra_str = Angle(d["RAJ2000"], "deg").to_string(unit="hour", precision=0)
        dec_str = Angle(d["DEJ2000"], "deg").to_string(unit="deg", precision=0)
        ss += "{:<20s} : {:8.3f} = {}\n".format("RA", d["RAJ2000"], ra_str)
        ss += "{:<20s} : {:8.3f} = {}\n".format("DEC", d["DEJ2000"], dec_str)

        ss += "{:<20s} : {:8.3f} +/- {:.3f} deg\n".format(
            "GLON", d["GLON"].value, d["GLON_Err"].value
        )
        ss += "{:<20s} : {:8.3f} +/- {:.3f} deg\n".format(
            "GLAT", d["GLAT"].value, d["GLAT_Err"].value
        )

        ss += "{:<20s} : {:.3f}\n".format("Position Error (68%)", d["Pos_Err_68"])
        ss += "{:<20s} : {:.3f}\n".format("Position Error (95%)", d["Pos_Err_95"])

        ss += "{:<20s} : {:.0f}\n".format("ROI number", d["ROI_Number"])
        ss += "{:<20s} : {}\n".format("Spatial model", d["Spatial_Model"])
        ss += "{:<20s} : {}\n".format("Spatial components", d["Components"])

        ss += "{:<20s} : {:.1f}\n".format("TS", d["Sqrt_TS"] ** 2)
        ss += "{:<20s} : {:.1f}\n".format("sqrt(TS)", d["Sqrt_TS"])

        ss += "{:<20s} : {:.3f} +/- {:.3f} (UL: {:.3f}) deg\n".format(
            "Size", d["Size"].value, d["Size_Err"].value, d["Size_UL"].value
        )

        ss += "{:<20s} : {:.3f}\n".format("R70", d["R70"])
        ss += "{:<20s} : {:.3f}\n".format("RSpec", d["RSpec"])

        ss += "{:<20s} : {:.1f}\n".format("Total model excess", d["Excess_Model_Total"])
        ss += "{:<20s} : {:.1f}\n".format("Excess in RSpec", d["Excess_RSpec"])
        ss += "{:<20s} : {:.1f}\n".format(
            "Model Excess in RSpec", d["Excess_RSpec_Model"]
        )
        ss += "{:<20s} : {:.1f}\n".format("Background in RSpec", d["Background_RSpec"])

        ss += "{:<20s} : {:.1f} hours\n".format("Livetime", d["Livetime"].value)

        ss += "{:<20s} : {:.2f}\n".format("Energy threshold", d["Energy_Threshold"])

        val, err = d["Flux_Map"].value, d["Flux_Map_Err"].value
        ss += "{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1 = ({:.2f} +/- {:.2f}) % Crab\n".format(  # noqa: 501
            "Source flux (>1 TeV)",
            val / FF,
            err / FF,
            val * FLUX_TO_CRAB,
            err * FLUX_TO_CRAB,
        )

        ss += "\nFluxes in RSpec (> 1 TeV):\n"

        ss += "{:<30s} : {:.3f} x 10^-12 cm^-2 s^-1 = {:5.2f} % Crab\n".format(
            "Map measurement",
            d["Flux_Map_RSpec_Data"].value / FF,
            d["Flux_Map_RSpec_Data"].value * FLUX_TO_CRAB,
        )

        ss += "{:<30s} : {:.3f} x 10^-12 cm^-2 s^-1 = {:5.2f} % Crab\n".format(
            "Source model",
            d["Flux_Map_RSpec_Source"].value / FF,
            d["Flux_Map_RSpec_Source"].value * FLUX_TO_CRAB,
        )

        ss += "{:<30s} : {:.3f} x 10^-12 cm^-2 s^-1 = {:5.2f} % Crab\n".format(
            "Other component model",
            d["Flux_Map_RSpec_Other"].value / FF,
            d["Flux_Map_RSpec_Other"].value * FLUX_TO_CRAB,
        )

        ss += "{:<30s} : {:.3f} x 10^-12 cm^-2 s^-1 = {:5.2f} % Crab\n".format(
            "Large scale component model",
            d["Flux_Map_RSpec_LS"].value / FF,
            d["Flux_Map_RSpec_LS"].value * FLUX_TO_CRAB,
        )

        ss += "{:<30s} : {:.3f} x 10^-12 cm^-2 s^-1 = {:5.2f} % Crab\n".format(
            "Total model",
            d["Flux_Map_RSpec_Total"].value / FF,
            d["Flux_Map_RSpec_Total"].value * FLUX_TO_CRAB,
        )

        ss += "{:<35s} : {:5.1f} %\n".format(
            "Containment in RSpec", 100 * d["Containment_RSpec"]
        )
        ss += "{:<35s} : {:5.1f} %\n".format(
            "Contamination in RSpec", 100 * d["Contamination_RSpec"]
        )
        label, val = (
            "Flux correction (RSpec -> Total)",
            100 * d["Flux_Correction_RSpec_To_Total"],
        )
        ss += f"{label:<35s} : {val:5.1f} %\n"
        label, val = (
            "Flux correction (Total -> RSpec)",
            100 * (1 / d["Flux_Correction_RSpec_To_Total"]),
        )
        ss += f"{label:<35s} : {val:5.1f} %\n"

        return ss

    def _info_spec(self):
        """Print information from spectral analysis."""
        d = self.data
        ss = "\n*** Info from spectral analysis ***\n\n"

        ss += "{:<20s} : {:.1f} hours\n".format("Livetime", d["Livetime_Spec"].value)

        lo = d["Energy_Range_Spec_Min"].value
        hi = d["Energy_Range_Spec_Max"].value
        ss += "{:<20s} : {:.2f} to {:.2f} TeV\n".format("Energy range:", lo, hi)

        ss += "{:<20s} : {:.1f}\n".format("Background", d["Background_Spec"])
        ss += "{:<20s} : {:.1f}\n".format("Excess", d["Excess_Spec"])
        ss += "{:<20s} : {}\n".format("Spectral model", d["Spectral_Model"])

        val = d["TS_ECPL_over_PL"]
        ss += "{:<20s} : {:.1f}\n".format("TS ECPL over PL", val)

        val = d["Flux_Spec_Int_1TeV"].value
        err = d["Flux_Spec_Int_1TeV_Err"].value
        ss += "{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1  = ({:.2f} +/- {:.2f}) % Crab\n".format(  # noqa: E501
            "Best-fit model flux(> 1 TeV)",
            val / FF,
            err / FF,
            val * FLUX_TO_CRAB,
            err * FLUX_TO_CRAB,
        )

        val = d["Flux_Spec_Energy_1_10_TeV"].value
        err = d["Flux_Spec_Energy_1_10_TeV_Err"].value
        ss += "{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 erg cm^-2 s^-1\n".format(
            "Best-fit model energy flux(1 to 10 TeV)", val / FF, err / FF
        )

        ss += self._info_spec_pl()
        ss += self._info_spec_ecpl()

        return ss

    def _info_spec_pl(self):
        d = self.data
        ss = "{:<20s} : {:.2f}\n".format("Pivot energy", d["Energy_Spec_PL_Pivot"])

        val = d["Flux_Spec_PL_Diff_Pivot"].value
        err = d["Flux_Spec_PL_Diff_Pivot_Err"].value
        ss += "{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1 TeV^-1  = ({:.2f} +/- {:.2f}) % Crab\n".format(  # noqa: E501
            "Flux at pivot energy",
            val / FF,
            err / FF,
            val * FLUX_TO_CRAB,
            err * FLUX_TO_CRAB_DIFF,
        )

        val = d["Flux_Spec_PL_Int_1TeV"].value
        err = d["Flux_Spec_PL_Int_1TeV_Err"].value
        ss += "{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1  = ({:.2f} +/- {:.2f}) % Crab\n".format(  # noqa: E501
            "PL   Flux(> 1 TeV)",
            val / FF,
            err / FF,
            val * FLUX_TO_CRAB,
            err * FLUX_TO_CRAB,
        )

        val = d["Flux_Spec_PL_Diff_1TeV"].value
        err = d["Flux_Spec_PL_Diff_1TeV_Err"].value
        ss += "{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1 TeV^-1  = ({:.2f} +/- {:.2f}) % Crab\n".format(  # noqa: E501
            "PL   Flux(@ 1 TeV)",
            val / FF,
            err / FF,
            val * FLUX_TO_CRAB,
            err * FLUX_TO_CRAB_DIFF,
        )

        val = d["Index_Spec_PL"]
        err = d["Index_Spec_PL_Err"]
        ss += "{:<20s} : {:.2f} +/- {:.2f}\n".format("PL   Index", val, err)

        return ss

    def _info_spec_ecpl(self):
        d = self.data
        ss = ""
        # ss = '{:<20s} : {:.1f}\n'.format('Pivot energy', d['Energy_Spec_ECPL_Pivot'])

        val = d["Flux_Spec_ECPL_Diff_1TeV"].value
        err = d["Flux_Spec_ECPL_Diff_1TeV_Err"].value
        ss += "{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1 TeV^-1  = ({:.2f} +/- {:.2f}) % Crab\n".format(  # noqa: E501
            "ECPL   Flux(@ 1 TeV)",
            val / FF,
            err / FF,
            val * FLUX_TO_CRAB,
            err * FLUX_TO_CRAB_DIFF,
        )

        val = d["Flux_Spec_ECPL_Int_1TeV"].value
        err = d["Flux_Spec_ECPL_Int_1TeV_Err"].value
        ss += "{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1  = ({:.2f} +/- {:.2f}) % Crab\n".format(  # noqa: E501
            "ECPL   Flux(> 1 TeV)",
            val / FF,
            err / FF,
            val * FLUX_TO_CRAB,
            err * FLUX_TO_CRAB,
        )

        val = d["Index_Spec_ECPL"]
        err = d["Index_Spec_ECPL_Err"]
        ss += "{:<20s} : {:.2f} +/- {:.2f}\n".format("ECPL Index", val, err)

        val = d["Lambda_Spec_ECPL"].value
        err = d["Lambda_Spec_ECPL_Err"].value
        ss += "{:<20s} : {:.3f} +/- {:.3f} TeV^-1\n".format("ECPL Lambda", val, err)

        # Use Gaussian analytical error propagation,
        # tested against the uncertainties package
        err = err / val**2
        val = 1.0 / val

        ss += "{:<20s} : {:.2f} +/- {:.2f} TeV\n".format("ECPL E_cut", val, err)

        return ss

    def _info_flux_points(self):
        """Print flux point results."""
        d = self.data
        ss = "\n*** Flux points info ***\n\n"
        ss += "Number of flux points: {}\n".format(d["N_Flux_Points"])
        ss += "Flux points table: \n\n"
        lines = format_flux_points_table(self.flux_points_table).pformat(
            max_width=-1, max_lines=-1
        )
        ss += "\n".join(lines)
        return ss + "\n"

    def _info_components(self):
        """Print information about the components."""
        ss = "\n*** Gaussian component info ***\n\n"
        ss += "Number of components: {}\n".format(len(self.components))
        ss += "{:<20s} : {}\n\n".format("Spatial components", self.data["Components"])

        for component in self.components:
            ss += str(component)
            ss += "\n\n"

        return ss

    @property
    def energy_range(self):
        """Spectral model energy range as a `~astropy.units.Quantity` with length 2."""
        energy_min, energy_max = (
            self.data["Energy_Range_Spec_Min"],
            self.data["Energy_Range_Spec_Max"],
        )

        if np.isnan(energy_min):
            energy_min = u.Quantity(0.2, "TeV")

        if np.isnan(energy_max):
            energy_max = u.Quantity(50, "TeV")

        return u.Quantity([energy_min, energy_max], "TeV")

    def spectral_model(self, which="best"):
        """Spectral model as a `~gammapy.modeling.models.SpectralModel` object.

        One of the following models (given by ``Spectral_Model`` in the catalog):

        - ``PL`` : `~gammapy.modeling.models.PowerLawSpectralModel`
        - ``ECPL`` : `~gammapy.modeling.models.ExpCutoffPowerLawSpectralModel`

        Parameters
        ----------
        which : {'best', 'pl', 'ecpl'}
            Which spectral model.
        """
        data = self.data

        if which == "best":
            spec_type = self.data["Spectral_Model"].strip().lower()
        elif which in {"pl", "ecpl"}:
            spec_type = which
        else:
            raise ValueError(f"Invalid selection: which = {which!r}")

        if spec_type == "pl":
            tag = "PowerLawSpectralModel"
            pars = {
                "index": data["Index_Spec_PL"],
                "amplitude": data["Flux_Spec_PL_Diff_Pivot"],
                "reference": data["Energy_Spec_PL_Pivot"],
            }
            errs = {
                "amplitude": data["Flux_Spec_PL_Diff_Pivot_Err"],
                "index": data["Index_Spec_PL_Err"],
            }
        elif spec_type == "ecpl":
            tag = "ExpCutoffPowerLawSpectralModel"
            pars = {
                "index": data["Index_Spec_ECPL"],
                "amplitude": data["Flux_Spec_ECPL_Diff_Pivot"],
                "reference": data["Energy_Spec_ECPL_Pivot"],
                "lambda_": data["Lambda_Spec_ECPL"],
            }
            errs = {
                "index": data["Index_Spec_ECPL_Err"],
                "amplitude": data["Flux_Spec_ECPL_Diff_Pivot_Err"],
                "lambda_": data["Lambda_Spec_ECPL_Err"],
            }
        else:
            raise ValueError(f"Invalid spec_type: {spec_type}")

        model = Model.create(tag, "spectral", **pars)
        errs["reference"] = 0 * u.TeV

        for name, value in errs.items():
            model.parameters[name].error = value

        return model

    def spatial_model(self):
        """Spatial model as a `~gammapy.modeling.models.SpatialModel` object.

        One of the following models (given by ``Spatial_Model`` in the catalog):

        - ``Point-Like`` or has a size upper limit : `~gammapy.modeling.models.PointSpatialModel`
        - ``Gaussian``: `~gammapy.modeling.models.GaussianSpatialModel`
        - ``2-Gaussian`` or ``3-Gaussian``: composite model (using ``+`` with Gaussians)
        - ``Shell``: `~gammapy.modeling.models.ShellSpatialModel`
        """
        d = self.data
        pars = {"lon_0": d["GLON"], "lat_0": d["GLAT"], "frame": "galactic"}
        errs = {"lon_0": d["GLON_Err"], "lat_0": d["GLAT_Err"]}

        spatial_type = self.data["Spatial_Model"]

        if spatial_type == "Point-Like":
            tag = "PointSpatialModel"
        elif spatial_type == "Gaussian":
            tag = "GaussianSpatialModel"
            pars["sigma"] = d["Size"]
            errs["sigma"] = d["Size_Err"]
        elif spatial_type in {"2-Gaussian", "3-Gaussian"}:
            raise ValueError("For Gaussian or Multi-Gaussian models, use sky_model()!")
        elif spatial_type == "Shell":
            # HGPS contains no information on shell width
            # Here we assume a 5% shell width for all shells.
            tag = "ShellSpatialModel"
            pars["radius"] = 0.95 * d["Size"]
            pars["width"] = d["Size"] - pars["radius"]
            errs["radius"] = d["Size_Err"]
        else:
            raise ValueError(f"Invalid spatial_type: {spatial_type}")

        model = Model.create(tag, "spatial", **pars)
        for name, value in errs.items():
            model.parameters[name].error = value
        return model

    def sky_model(self, which="best"):
        """Source sky model.

        Parameters
        ----------
        which : {'best', 'pl', 'ecpl'}
            Which spectral model.

        Returns
        -------
        sky_model : `~gammapy.modeling.models.Models`
           Models of the catalog object.
        """
        models = self.components_models(which=which)
        if len(models) > 1:
            geom = self._get_components_geom(models)
            return models.to_template_sky_model(geom=geom, name=self.name)
        else:
            return models[0]

    def components_models(self, which="best", linked=False):
        """Models of the source components.

        Parameters
        ----------
        which : {'best', 'pl', 'ecpl'}
            Which spectral model.

        linked : bool
             Each subcomponent of a source is given as a different `SkyModel`.
             If True, the spectral parameters except the normalisation are linked.
             Default is False.

        Returns
        -------
        sky_model : `~gammapy.modeling.models.Models`
           Models of the catalog object.
        """
        spatial_type = self.data["Spatial_Model"]
        missing_size = (
            spatial_type == "Gaussian" and self.spatial_model().sigma.value == 0
        )
        if spatial_type in {"2-Gaussian", "3-Gaussian"} or missing_size:
            models = []
            spectral_model = self.spectral_model(which=which)
            for component in self.components:
                spec_component = spectral_model.copy()
                weight = component.data["Flux_Map"] / self.data["Flux_Map"]
                spec_component.parameters["amplitude"].value *= weight
                if linked:
                    for name in spec_component.parameters.names:
                        if name not in ["norm", "amplitude"]:
                            spec_component.__dict__[name] = spectral_model.parameters[
                                name
                            ]
                model = SkyModel(
                    spatial_model=component.spatial_model(),
                    spectral_model=spec_component,
                    name=component.name,
                )
                models.append(model)
        else:
            models = [
                SkyModel(
                    spatial_model=self.spatial_model(),
                    spectral_model=self.spectral_model(which=which),
                    name=self.name,
                )
            ]
        return Models(models)

    @staticmethod
    def _get_components_geom(models):
        energy_axis = MapAxis.from_energy_bounds(
            "100 GeV", "100 TeV", nbin=10, per_decade=True, name="energy_true"
        )
        regions = [m.spatial_model.evaluation_region for m in models]
        geom = RegionGeom.from_regions(
            regions, binsz_wcs="0.02 deg", axes=[energy_axis]
        )
        return geom.to_wcs_geom()

    @property
    def flux_points_table(self):
        """Flux points table as a `~astropy.table.Table`."""
        table = Table()
        table.meta["sed_type_init"] = "dnde"
        table.meta["n_sigma_ul"] = 2
        table.meta["n_sigma"] = 1
        table.meta["sqrt_ts_threshold_ul"] = 1
        mask = ~np.isnan(self.data["Flux_Points_Energy"])

        table["e_ref"] = self.data["Flux_Points_Energy"][mask]
        table["e_min"] = self.data["Flux_Points_Energy_Min"][mask]
        table["e_max"] = self.data["Flux_Points_Energy_Max"][mask]

        table["dnde"] = self.data["Flux_Points_Flux"][mask]
        table["dnde_errn"] = self.data["Flux_Points_Flux_Err_Lo"][mask]
        table["dnde_errp"] = self.data["Flux_Points_Flux_Err_Hi"][mask]
        table["dnde_ul"] = self.data["Flux_Points_Flux_UL"][mask]
        table["is_ul"] = self.data["Flux_Points_Flux_Is_UL"][mask].astype("bool")
        return table


class SourceCatalogHGPS(SourceCatalog):
    """HESS Galactic plane survey (HGPS) source catalog.

    Reference: https://www.mpi-hd.mpg.de/hfm/HESS/hgps/

    One source is represented by `SourceCatalogObjectHGPS`.

    Examples
    --------
    Let's assume you have downloaded the HGPS catalog FITS file, e.g. via:

    .. code-block:: bash

        curl -O https://www.mpi-hd.mpg.de/hfm/HESS/hgps/data/hgps_catalog_v1.fits.gz

    Then you can load it like this:

    >>> import matplotlib.pyplot as plt
    >>> from gammapy.catalog import SourceCatalogHGPS
    >>> filename = '$GAMMAPY_DATA/catalogs/hgps_catalog_v1.fits.gz'
    >>> cat = SourceCatalogHGPS(filename)

    Access a source by name:

    >>> source = cat['HESS J1843-033']
    >>> print(source)
    <BLANKLINE>
    *** Basic info ***
    <BLANKLINE>
    Catalog row index (zero-based) : 64
    Source name          : HESS J1843-033
    Analysis reference   : HGPS
    Source class         : Unid
    Identified object    : --
    Gamma-Cat id         : 126
    <BLANKLINE>
    <BLANKLINE>
    *** Info from map analysis ***
    <BLANKLINE>
    RA                   :  280.952 deg = 18h43m48s
    DEC                  :   -3.554 deg = -3d33m15s
    GLON                 :   28.899 +/- 0.072 deg
    GLAT                 :    0.075 +/- 0.036 deg
    Position Error (68%) : 0.122 deg
    Position Error (95%) : 0.197 deg
    ROI number           : 3
    Spatial model        : 2-Gaussian
    Spatial components   : HGPSC 083, HGPSC 084
    TS                   : 256.6
    sqrt(TS)             : 16.0
    Size                 : 0.239 +/- 0.063 (UL: 0.000) deg
    R70                  : 0.376 deg
    RSpec                : 0.376 deg
    Total model excess   : 979.5
    Excess in RSpec      : 775.6
    Model Excess in RSpec : 690.2
    Background in RSpec  : 1742.4
    Livetime             : 41.8 hours
    Energy threshold     : 0.38 TeV
    Source flux (>1 TeV) : (2.882 +/- 0.305) x 10^-12 cm^-2 s^-1 = (12.75 +/- 1.35) % Crab
    <BLANKLINE>
    Fluxes in RSpec (> 1 TeV):
    Map measurement                : 2.267 x 10^-12 cm^-2 s^-1 = 10.03 % Crab
    Source model                   : 2.018 x 10^-12 cm^-2 s^-1 =  8.93 % Crab
    Other component model          : 0.004 x 10^-12 cm^-2 s^-1 =  0.02 % Crab
    Large scale component model    : 0.361 x 10^-12 cm^-2 s^-1 =  1.60 % Crab
    Total model                    : 2.383 x 10^-12 cm^-2 s^-1 = 10.54 % Crab
    Containment in RSpec                :  70.0 %
    Contamination in RSpec              :  15.3 %
    Flux correction (RSpec -> Total)    : 121.0 %
    Flux correction (Total -> RSpec)    :  82.7 %
    <BLANKLINE>
    *** Info from spectral analysis ***
    <BLANKLINE>
    Livetime             : 16.8 hours
    Energy range:        : 0.22 to 61.90 TeV
    Background           : 5126.9
    Excess               : 980.1
    Spectral model       : PL
    TS ECPL over PL      : --
    Best-fit model flux(> 1 TeV) : (3.043 +/- 0.196) x 10^-12 cm^-2 s^-1  = (13.47 +/- 0.87) % Crab
    Best-fit model energy flux(1 to 10 TeV) : (10.918 +/- 0.733) x 10^-12 erg cm^-2 s^-1
    Pivot energy         : 1.87 TeV
    Flux at pivot energy : (0.914 +/- 0.058) x 10^-12 cm^-2 s^-1 TeV^-1  = (4.04 +/- 0.17) % Crab
    PL   Flux(> 1 TeV)   : (3.043 +/- 0.196) x 10^-12 cm^-2 s^-1  = (13.47 +/- 0.87) % Crab
    PL   Flux(@ 1 TeV)   : (3.505 +/- 0.247) x 10^-12 cm^-2 s^-1 TeV^-1  = (15.51 +/- 0.70) % Crab
    PL   Index           : 2.15 +/- 0.05
    ECPL   Flux(@ 1 TeV) : (0.000 +/- 0.000) x 10^-12 cm^-2 s^-1 TeV^-1  = (0.00 +/- 0.00) % Crab
    ECPL   Flux(> 1 TeV) : (0.000 +/- 0.000) x 10^-12 cm^-2 s^-1  = (0.00 +/- 0.00) % Crab
    ECPL Index           : -- +/- --
    ECPL Lambda          : 0.000 +/- 0.000 TeV^-1
    ECPL E_cut           : inf +/- nan TeV
    <BLANKLINE>
    *** Flux points info ***
    <BLANKLINE>
    Number of flux points: 6
    Flux points table:
    <BLANKLINE>
    e_ref  e_min  e_max        dnde         dnde_errn       dnde_errp        dnde_ul     is_ul
     TeV    TeV    TeV   1 / (TeV s cm2) 1 / (TeV s cm2) 1 / (TeV s cm2) 1 / (TeV s cm2)
    ------ ------ ------ --------------- --------------- --------------- --------------- -----
     0.332  0.215  0.511       3.048e-11       6.890e-12       7.010e-12       4.455e-11 False
     0.787  0.511  1.212       5.383e-12       6.655e-13       6.843e-13       6.739e-12 False
     1.957  1.212  3.162       9.160e-13       9.732e-14       1.002e-13       1.120e-12 False
     4.870  3.162  7.499       1.630e-13       2.001e-14       2.097e-14       2.054e-13 False
    12.115  7.499 19.573       1.648e-14       3.124e-15       3.348e-15       2.344e-14 False
    30.142 19.573 46.416       7.777e-16       4.468e-16       5.116e-16       1.883e-15 False
    <BLANKLINE>
    *** Gaussian component info ***
    <BLANKLINE>
    Number of components: 2
    Spatial components   : HGPSC 083, HGPSC 084
    <BLANKLINE>
    Component HGPSC 083:
    GLON                 :   29.047 +/- 0.026 deg
    GLAT                 :    0.244 +/- 0.027 deg
    Size                 : 0.125 +/- 0.021 deg
    Flux (>1 TeV)        : (1.34 +/- 0.36) x 10^-12 cm^-2 s^-1 = (5.9 +/- 1.6) % Crab
    <BLANKLINE>
    Component HGPSC 084:
    GLON                 :   28.770 +/- 0.059 deg
    GLAT                 :   -0.073 +/- 0.069 deg
    Size                 : 0.229 +/- 0.046 deg
    Flux (>1 TeV)        : (1.54 +/- 0.47) x 10^-12 cm^-2 s^-1 = (6.8 +/- 2.1) % Crab
    <BLANKLINE>
    <BLANKLINE>
    *** Source associations info ***
    <BLANKLINE>
      Source_Name    Association_Catalog    Association_Name   Separation
                                                                  deg
    ---------------- ------------------- --------------------- ----------
      HESS J1843-033                3FGL     3FGL J1843.7-0322   0.178442
      HESS J1843-033                3FGL     3FGL J1844.3-0344   0.242835
      HESS J1843-033                 SNR             G28.6-0.1   0.330376


    Access source spectral data and plot it:

    >>> ax = plt.subplot()
    >>> source.spectral_model().plot(source.energy_range, ax=ax) # doctest: +SKIP
    >>> source.spectral_model().plot_error(source.energy_range, ax=ax) # doctest: +SKIP
    >>> source.flux_points.plot(ax=ax) # doctest: +SKIP

    Gaussian component information can be accessed as well,
    either via the source, or via the catalog:

    >>> source.components
    [SourceCatalogObjectHGPSComponent('HGPSC 083'), SourceCatalogObjectHGPSComponent('HGPSC 084')]

    >>> cat.gaussian_component(83)
    SourceCatalogObjectHGPSComponent('HGPSC 084')
    """

    tag = "hgps"
    """Source catalog name (str)."""

    description = "H.E.S.S. Galactic plane survey (HGPS) source catalog"
    """Source catalog description (str)."""

    source_object_class = SourceCatalogObjectHGPS

    def __init__(
        self,
        filename="$GAMMAPY_DATA/catalogs/hgps_catalog_v1.fits.gz",
        hdu="HGPS_SOURCES",
    ):
        filename = make_path(filename)
        table = Table.read(filename, hdu=hdu)

        source_name_alias = ("Identified_Object",)
        super().__init__(table=table, source_name_alias=source_name_alias)

        self._table_components = Table.read(filename, hdu="HGPS_GAUSS_COMPONENTS")
        self._table_associations = Table.read(filename, hdu="HGPS_ASSOCIATIONS")
        self._table_associations["Separation"].format = ".6f"
        self._table_identifications = Table.read(filename, hdu="HGPS_IDENTIFICATIONS")
        self._table_large_scale_component = Table.read(
            filename, hdu="HGPS_LARGE_SCALE_COMPONENT"
        )

    @property
    def table_components(self):
        """Gaussian component table as a `~astropy.table.Table`."""
        return self._table_components

    @property
    def table_associations(self):
        """Source association table as a `~astropy.table.Table`."""
        return self._table_associations

    @property
    def table_identifications(self):
        """Source identification table as a `~astropy.table.Table`."""
        return self._table_identifications

    @property
    def table_large_scale_component(self):
        """Large scale component table as a `~astropy.table.Table`."""
        return self._table_large_scale_component

    @property
    def large_scale_component(self):
        """Large scale component model as a `~gammapy.catalog.SourceCatalogLargeScaleHGPS` object."""
        return SourceCatalogLargeScaleHGPS(self.table_large_scale_component)

    def _make_source_object(self, index):
        """Make `SourceCatalogObject` for given row index."""
        source = super()._make_source_object(index)

        if source.data["Components"] not in ["", "--"]:
            source.components = list(self._get_gaussian_components(source))

        self._attach_association_info(source)

        if source.data["Source_Class"] != "Unid":
            self._attach_identification_info(source)

        return source

    def _get_gaussian_components(self, source):
        for name in source.data["Components"].split(", "):
            row_index = int(name.split()[-1]) - 1
            yield self.gaussian_component(row_index)

    def _attach_association_info(self, source):
        t = self.table_associations
        mask = source.data["Source_Name"] == t["Source_Name"]
        source.associations = t[mask]

    def _attach_identification_info(self, source):
        t = self._table_identifications
        idx = np.nonzero(source.name == t["Source_Name"])[0][0]
        source.identification_info = table_row_to_dict(t[idx])

    def gaussian_component(self, row_idx):
        """Gaussian component as a `SourceCatalogObjectHGPSComponent` object."""
        data = table_row_to_dict(self.table_components[row_idx])
        data[SourceCatalogObject._row_index_key] = row_idx
        return SourceCatalogObjectHGPSComponent(data=data)

    def to_models(self, which="best", components_status="independent"):
        """Create Models object from catalog.

        Parameters
        ----------
        which : {'best', 'pl', 'ecpl'}
            Which spectral model.

        components_status : {'independent', 'linked', 'merged'}
            Relation between the sources components. Available options are:
                * 'independent': each subcomponent of a source is given as a different `SkyModel` (Default).
                * 'linked': each subcomponent of a source is given as a different `SkyModel` but the spectral
                  parameters except the normalisation are linked.
                * 'merged': the subcomponents are merged into a single `SkyModel` given as a
                  `~gammapy.modeling.models.TemplateSpatialModel` with a `~gammapy.modeling.models.PowerLawNormSpectralModel`.
                  In that case the relative weights between the components cannot be adjusted.

        Returns
        -------
        models : `~gammapy.modeling.models.Models`
            Models of the catalog.
        """
        models = []
        for source in self:
            if components_status == "merged":
                m = [source.sky_model(which=which)]
            else:
                m = source.components_models(
                    which=which, linked=components_status == "linked"
                )
            models.extend(m)
        return Models(models)


class SourceCatalogLargeScaleHGPS:
    """Gaussian band model.

    This 2-dimensional model is Gaussian in ``y`` for a given ``x``,
    and the Gaussian parameters can vary in ``x``.

    One application of this model is the diffuse emission along the
    Galactic plane, i.e. ``x = GLON`` and ``y = GLAT``.

    Parameters
    ----------
    table : `~astropy.table.Table`
        Table of Gaussian parameters.
        ``x``, ``amplitude``, ``mean``, ``stddev``.
    interp_kwargs : dict
        Keyword arguments passed to `ScaledRegularGridInterpolator`.
    """

    def __init__(self, table, interp_kwargs=None):
        interp_kwargs = interp_kwargs or {}
        interp_kwargs.setdefault("values_scale", "lin")

        self.table = table
        glon = Angle(self.table["GLON"]).wrap_at("180d")

        interps = {}

        for column in table.colnames:
            values = self.table[column].quantity
            interp = ScaledRegularGridInterpolator((glon,), values, **interp_kwargs)
            interps[column] = interp

        self._interp = interps

    def _interpolate_parameter(self, parname, glon):
        glon = glon.wrap_at("180d")
        return self._interp[parname]((np.asanyarray(glon),), clip=False)

    def peak_brightness(self, glon):
        """Peak brightness at a given longitude.

        Parameters
        ----------
        glon : `~astropy.coordinates.Angle`
            Galactic Longitude.
        """
        return self._interpolate_parameter("Surface_Brightness", glon)

    def peak_brightness_error(self, glon):
        """Peak brightness error at a given longitude.

        Parameters
        ----------
        glon : `~astropy.coordinates.Angle`
            Galactic Longitude.
        """
        return self._interpolate_parameter("Surface_Brightness_Err", glon)

    def width(self, glon):
        """Width at a given longitude.

        Parameters
        ----------
        glon : `~astropy.coordinates.Angle`
            Galactic Longitude.
        """
        return self._interpolate_parameter("Width", glon)

    def width_error(self, glon):
        """Width error at a given longitude.

        Parameters
        ----------
        glon : `~astropy.coordinates.Angle`
            Galactic Longitude.
        """
        return self._interpolate_parameter("Width_Err", glon)

    def peak_latitude(self, glon):
        """Peak position at a given longitude.

        Parameters
        ----------
        glon : `~astropy.coordinates.Angle`
            Galactic Longitude.
        """
        return self._interpolate_parameter("GLAT", glon)

    def peak_latitude_error(self, glon):
        """Peak position error at a given longitude.

        Parameters
        ----------
        glon : `~astropy.coordinates.Angle`
            Galactic Longitude.
        """
        return self._interpolate_parameter("GLAT_Err", glon)

    def evaluate(self, position):
        """Evaluate model at a given position.

        Parameters
        ----------
        position : `~astropy.coordinates.SkyCoord`
            Position on the sky.
        """
        glon, glat = position.galactic.l, position.galactic.b
        width = self.width(glon)
        amplitude = self.peak_brightness(glon)
        mean = self.peak_latitude(glon)
        return Gaussian1D.evaluate(glat, amplitude=amplitude, mean=mean, stddev=width)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utilities to find roots of a scalar function within a given range."""

import numpy as np
from scipy.optimize import RootResults, root_scalar
import astropy.units as u
from gammapy.utils.interpolation import interpolation_scale

try:
    BAD_RES = RootResults(
        root=np.nan, iterations=0, function_calls=0, flag=0, method="brentq"
    )
except TypeError:
    BAD_RES = RootResults(root=np.nan, iterations=0, function_calls=0, flag=0)


def find_roots(
    f,
    lower_bound,
    upper_bound,
    nbin=100,
    points_scale="lin",
    args=(),
    method="brentq",
    fprime=None,
    fprime2=None,
    xtol=None,
    rtol=None,
    maxiter=None,
    options=None,
):
    """Find roots of a scalar function within a given range.

    Parameters
    ----------
    f : callable
        A function to find roots of. Its output should be unitless.
    lower_bound : `~astropy.units.Quantity`
        Lower bound of the search ranges to find roots.
        If an array is given search will be performed element-wise.
    upper_bound : `~astropy.units.Quantity`
        Upper bound of the search ranges to find roots.
        If an array is given search will be performed element-wise.
    nbin : int, optional
        Number of bins to sample the search range, ignored if bounds are arrays.
        Default is 100.
    points_scale : {"lin", "log", "sqrt"}, optional
        Scale used to sample the search range. Default is "lin".
    args : tuple, optional
        Extra arguments passed to the objective function and its derivative(s).
    method : str, optional
        Solver available in `~scipy.optimize.root_scalar`.  Should be one of :
            - 'brentq' (default),
            - 'brenth',
            - 'bisect',
            - 'ridder',
            - 'toms748',
            - 'newton',
            - 'secant',
            - 'halley',
    fprime : bool or callable, optional
        If `fprime` is a boolean and is True, `f` is assumed to return the
        value of the objective function and of the derivative.
        `fprime` can also be a callable returning the derivative of `f`. In
        this case, it must accept the same arguments as `f`.
    fprime2 : bool or callable, optional
        If `fprime2` is a boolean and is True, `f` is assumed to return the
        value of the objective function and of the
        first and second derivatives.
        `fprime2` can also be a callable returning the second derivative of `f`.
        In this case, it must accept the same arguments as `f`.
    xtol : float, optional
        Tolerance (absolute) for termination.
    rtol : float, optional
        Tolerance (relative) for termination.
    maxiter : int, optional
        Maximum number of iterations.
    options : dict, optional
        A dictionary of solver options.
        See `~scipy.optimize.root_scalar` for details.

    Returns
    -------
    roots : `~astropy.units.Quantity`
        The function roots.

    results : `~numpy.array`
        An array of `~scipy.optimize.RootResults` which is an
        object containing information about the convergence.
        If the solver failed to converge in a bracketing range
        the corresponding `roots` array element is NaN.

    """
    kwargs = dict(
        args=args,
        method=method,
        fprime=fprime,
        fprime2=fprime2,
        xtol=xtol,
        rtol=rtol,
        maxiter=maxiter,
        options=options,
    )

    if isinstance(lower_bound, u.Quantity):
        unit = lower_bound.unit
        lower_bound = lower_bound.value
        upper_bound = u.Quantity(upper_bound).to_value(unit)
    else:
        unit = 1

    scale = interpolation_scale(points_scale)
    a = scale(lower_bound)
    b = scale(upper_bound)
    x = scale.inverse(np.linspace(a, b, nbin + 1))
    if len(x) > 2:
        signs = np.sign([f(xk, *args) for xk in x])
        ind = np.where(signs[:-1] != signs[1:])[0]
    else:
        ind = [0]
    nroots = max(1, len(ind))
    roots = np.ones(nroots) * np.nan
    results = np.array(nroots * [BAD_RES])

    for k, idx in enumerate(ind):
        bracket = [x[idx], x[idx + 1]]
        if method in ["bisection", "brentq", "brenth", "ridder", "toms748"]:
            kwargs["bracket"] = bracket
        elif method in ["secant", "newton", "halley"]:
            kwargs["x0"] = bracket[0]
            kwargs["x1"] = bracket[1]
        else:
            raise ValueError(f'Unknown solver "{method}"')
        try:
            res = root_scalar(f, **kwargs)
            results[k] = res
            if res.converged:
                roots[k] = res.root
        except (RuntimeError, ValueError):
            continue
    return roots * unit, results



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import astropy.units as u
from astropy.table import Column, Table
from gammapy.utils.table import table_row_to_dict, table_standardise_units_copy


def test_table_standardise_units():
    table = Table(
        [
            Column([1], "a", unit="ph cm-2 s-1"),
            Column([1], "b", unit="ct cm-2 s-1"),
            Column([1], "c", unit="cm-2 s-1"),
            Column([1], "d"),
        ]
    )

    table = table_standardise_units_copy(table)

    assert table["a"].unit == "cm-2 s-1"
    assert table["b"].unit == "cm-2 s-1"
    assert table["c"].unit == "cm-2 s-1"
    assert table["d"].unit is None


@pytest.fixture()
def table():
    return Table(
        [Column([1, 2], "a"), Column([1, 2] * u.m, "b"), Column(["x", "yy"], "c")]
    )


def test_table_row_to_dict(table):
    actual = table_row_to_dict(table[1])
    expected = {"a": 2, "b": 2 * u.m, "c": "yy"}
    assert actual == expected



"""
Mask maps
=========

Create and apply masks maps.

Prerequisites
-------------

-  Understanding of basic analyses in 1D or 3D.
-  Usage of `~regions` and catalogs, see the :doc:`catalog
   notebook </tutorials/details/catalog>`.

Context
-------

There are two main categories of masks in Gammapy for different use
cases. - Fitting often requires to ignore some parts of a reduced
dataset, e.g. to restrict the fit to a specific energy range or to
ignore parts of the region of interest that the user does not want to
model, or both. Gammapy’s `~gammapy.datasets.Datasets` therefore contain a ``mask_fit``
sharing the same geometry as the data (i.e. ``counts``). - During data
reduction, some background makers will normalize the background model
template on the data themselves. To limit contamination by real photons,
one has to exclude parts of the field-of-view where signal is expected
to be large. To do so, one needs to provide an exclusion mask. The
latter can be provided in a different geometry as it will be reprojected
by the `~gammapy.makers.Maker` class.

We explain in more details these two types of masks below:

Masks for fitting
~~~~~~~~~~~~~~~~~

The region of interest used for the fit can defined through the dataset
``mask_fit`` attribute. The ``mask_fit`` is a map containing boolean
values where pixels used in the fit are stored as True.

A spectral fit (1D or 3D) can be restricted to a specific energy range
where e.g. the background is well estimated or where the number of
counts is large enough. Similarly, 2D and 3D analyses usually require to
work with a wider map than the region of interest so sources laying
outside but reconstructed inside because of the PSF are correctly taken
into account. Then the ``mask_fit`` have to include a margin that take
into account the PSF width. We will show an example in the boundary mask
sub-section.

The ``mask_fit`` also can be used to exclude sources or complex regions
for which we don’t have good enough models. In that case the masking is
an extra security, it is preferable to include the available models
even if the sources are masked and frozen.

Note that a dataset contains also a ``mask_safe`` attribute that is
created and filled during data reduction. It is not to be modified
directly by users. The ``mask_safe`` is defined only from the options
passed to the `~gammapy.makers.SafeMaskMaker`.

Exclusion masks
~~~~~~~~~~~~~~~

Background templates stored in the DL3 IRF are often not reliable enough
to be used without some corrections. A set of common techniques to
perform background or normalisation from the data is implemented in
gammapy: reflected regions for 1D spectrum analysis, field-of-view (FoV)
background or ring background for 2D and 3D analyses.

To avoid contamination of the background estimate from gamma-ray bright
regions these methods require to exclude those regions from the data
used for the estimation. To do so, we use exclusion masks. They are maps
containing boolean values where excluded pixels are stored as False.

Proposed approach
-----------------

Even if the use cases for exclusion masks and fit masks are different,
the way to create these masks is exactly the same, so in the following
we show how to work with masks in general:

- Creating masks from scratch
- Combining multiple masks
- Extending and reducing an existing mask
- Reading and writing masks

"""

######################################################################
# Setup
# -----
#

import numpy as np
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
from regions import CircleSkyRegion, Regions

# %matplotlib inline
import matplotlib.pyplot as plt
from gammapy.catalog import CATALOG_REGISTRY
from gammapy.datasets import Datasets
from gammapy.estimators import ExcessMapEstimator
from gammapy.maps import Map, WcsGeom


######################################################################
# .. _masks-for-fitting:
#
# Creating a mask for fitting
# ---------------------------
#
# One can build a ``mask_fit`` to restrict the energy range of pixels used
# to fit a `~gammapy.datasets.Dataset`. As the mask is a `~gammapy.maps.Map` object,
# it must use the same geometry (i.e. a `~gammapy.maps.Geom` object) as the `~gammapy.datasets.Dataset`
# it is applied to.
#
# We show here how to proceed on a `~gammapy.datasets.MapDataset` taken from Fermi data
# used in the 3FHL catalog. The dataset is already in the form of a
# `~gammapy.datasets.Datasets` object, and we begin by reading it from disk.
#

filename = "$GAMMAPY_DATA/fermi-3fhl-crab/Fermi-LAT-3FHL_datasets.yaml"
datasets = Datasets.read(filename=filename)
dataset = datasets["Fermi-LAT"]


######################################################################
# We can check the default energy range of the dataset. In the absence of
# a ``mask_fit`` it is equal to the safe energy range.
#

print(f"Fit range : {dataset.energy_range_total}")


######################################################################
# Create a mask in energy
# ~~~~~~~~~~~~~~~~~~~~~~~
#
# We show first how to use a simple helper function
# `~gammapy.maps.Geom.energy_mask`.
#
# We obtain the `~gammapy.maps.Geom` that is stored on the ``counts`` map inside the
# `~gammapy.datasets.Dataset` and we can directly create the `~gammapy.maps.Map`.
#

mask_energy = dataset.counts.geom.energy_mask(10 * u.GeV, 700 * u.GeV)


######################################################################
# We can now set the dataset ``mask_fit`` attribute.
#
# And we check that the total fit range has changed accordingly. The bin
# edges closest to requested range provide the actual fit range.
#

dataset.mask_fit = mask_energy
print(f"Fit range : {dataset.energy_range_total}")


######################################################################
# Mask some sky regions
# ~~~~~~~~~~~~~~~~~~~~~
#
# One might also exclude some specific part of the sky for the fit. For
# instance, if one wants not to model a specific source in the region of
# interest, or if one want to reduce the region of interest in the dataset
# `~gammapy.maps.Geom`.
#
# In the following we restrict the fit region to a square around the Crab
# nebula. **Note**: the dataset geometry is aligned on the galactic frame,
# we use the same frame to define the box to ensure a correct alignment.
# We can now create the map. We use the `~gammapy.maps.WcsGeom.region_mask` method
# putting all pixels outside the regions to False (because we only want to
# consider pixels inside the region. For convenience, we can directly pass
# a ds9 region string to the method:
#

regions = "galactic;box(184.55, -5.78, 3.0, 3.0)"
mask_map = dataset.counts.geom.region_mask(regions)


######################################################################
# We can now combine this mask with the energy mask using the logical and
# operator
#

dataset.mask_fit &= mask_map


######################################################################
# Let’s check the result and plot the full mask.
#

dataset.mask_fit.plot_grid(ncols=5, vmin=0, vmax=1, figsize=(14, 3))
plt.show()


######################################################################
# Creating a mask manually
# ~~~~~~~~~~~~~~~~~~~~~~~~
#
# If you are more familiar with the `~gammapy.maps.Geom` and `~gammapy.maps.Map` API, you can also
# create the mask manually from the coordinates of all pixels in the
# geometry. Here we simply show how to obtain the same behaviour as the
# `~gammapy.maps.Geom.energy_mask` helper method.
#
# In practice, this allows to create complex energy dependent masks.
#

coords = dataset.counts.geom.get_coord()
mask_data = (coords["energy"] >= 10 * u.GeV) & (coords["energy"] < 700 * u.GeV)
mask_energy = Map.from_geom(dataset.counts.geom, data=mask_data)


######################################################################
# Creating an exclusion mask
# --------------------------
#
# Exclusion masks are typically used for background estimation to mask out
# regions where gamma-ray signal is expected. An exclusion mask is usually
# a simple 2D boolean `~gammapy.maps.Map` where excluded positions are stored as
# `False`. Their actual geometries are independent of the target
# datasets that a user might want to build. The first thing to do is to
# build the geometry.
#
# Define the geometry
# ~~~~~~~~~~~~~~~~~~~
#
# Masks are stored in `~gammapy.maps.Map` objects. We must first define its geometry
# and then we can determine which pixels to exclude. Here we consider a
# region at the Galactic anti-centre around the crab nebula.
#

position = SkyCoord(83.633083, 22.0145, unit="deg", frame="icrs")
geom = WcsGeom.create(skydir=position, width="5 deg", binsz=0.02, frame="galactic")


######################################################################
# Create the mask from a list of regions
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# One can build an exclusion mask from regions. We show here how to
# proceed.
#
# We can rely on known sources positions and properties to build a list of
# regions (here `~regions.SkyRegion`) enclosing most of the signal that
# our detector would see from these objects.
#
# A useful function to create region objects is
# `~regions.Regions.parse`. It can take strings defining regions
# e.g. following the “ds9” format and convert them to `regions`.
#
# Here we use a region enclosing the Crab nebula with 0.3 degrees. The
# actual region size should depend on the expected PSF of the data used.
# We also add another region with a different shape as en example.
#

regions_ds9 = "galactic;box(185,-4,1.0,0.5, 45);icrs;circle(83.633083, 22.0145, 0.3)"
regions = Regions.parse(regions_ds9, format="ds9")
print(regions)


######################################################################
# Equivalently the regions can be read from a ds9 file, this time using
# `~regions.Regions.read`.
#

# regions = Regions.read('ds9.reg', format="ds9")


######################################################################
# Create the mask map
# ^^^^^^^^^^^^^^^^^^^
#
# We can now create the map. We use the `~gammapy.maps.WcsGeom.region_mask` method
# putting all pixels inside the regions to False.
#

# to define the exclusion mask we take the inverse
mask_map = ~geom.region_mask(regions)
mask_map.plot()
plt.show()


######################################################################
# Create the mask from a catalog of sources
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# We can also build our list of regions from a list of catalog sources.
# Here we use the Fermi 4FGL catalog which we read using
# `~gammapy.catalog.SourceCatalog`.
#

fgl = CATALOG_REGISTRY.get_cls("4fgl")()


######################################################################
# We now select sources that are contained in the region we are interested
# in.
#

inside_geom = geom.contains(fgl.positions)
positions = fgl.positions[inside_geom]


######################################################################
# We now create the list of regions using our 0.3 degree radius a priori
# value. If the sources were extended, one would have to adapt the sizes
# to account for the larger size.
#

exclusion_radius = Angle("0.3 deg")
regions = [CircleSkyRegion(position, exclusion_radius) for position in positions]


######################################################################
# Now we can build the mask map the same way as above.
#

mask_map_catalog = ~geom.region_mask(regions)
mask_map_catalog.plot()
plt.show()


######################################################################
# Create the mask from statistically significant pixels in a dataset
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Here we want to determine an exclusion from the data directly. We will
# estimate the significance of the data using the `~gammapy.estimators.ExcessMapEstimator`,
# and exclude all pixels above a given threshold.
#
# Here we use the `~gammapy.datasets.MapDataset` taken from the Fermi data used above.
#


######################################################################
# We apply a significance estimation. We integrate the counts using a
# correlation radius of 0.4 degree and apply regular significance
# estimate.
#

estimator = ExcessMapEstimator("0.4 deg", selection_optional=[])
result = estimator.run(dataset)


######################################################################
# Finally, we create the mask map by applying a threshold of 5 sigma to
# remove pixels.
#

significance_mask = result["sqrt_ts"] < 5.0


######################################################################
# Because the `~gammapy.estimators.ExcessMapEstimator` returns NaN for masked pixels, we
# need to put the NaN values to `True` to avoid incorrectly excluding
# them.
#

invalid_pixels = np.isnan(result["sqrt_ts"].data)
significance_mask.data[invalid_pixels] = True
significance_mask.plot()
plt.show()


######################################################################
# This method frequently yields isolated pixels or weakly significant
# features if one places the threshold too low.
#
# To overcome this issue, one can use
# `~skimage.filters.apply_hysteresis_threshold` . This filter allows to
# define two thresholds and mask only the pixels between the low and high
# thresholds if they are not continuously connected to a pixel above the
# high threshold. This allows to better preserve the structure of the
# excesses.
#
# Note that scikit-image is not a required dependency of gammapy, you
# might need to install it.
#


######################################################################
# Masks operations
# ----------------
#
# If two masks share the same geometry it is easy to combine them with
# `~gammapy.maps.Map` arithmetic.
#
# OR condition is represented by ``|`` operator :
#

mask = mask_map | mask_map_catalog
mask.plot()
plt.show()


######################################################################
# AND condition is represented by ``&`` or ``*`` operators :
#

mask_map &= mask_map_catalog
mask_map.plot()
plt.show()


######################################################################
# The NOT operator is represented by the ``~`` symbol:
#

significance_mask_inv = ~significance_mask
significance_mask_inv.plot()
plt.show()


######################################################################
# Mask modifications
# ------------------
#
# Mask dilation and erosion
# ~~~~~~~~~~~~~~~~~~~~~~~~~
#
# One can reduce or extend a mask using `~gammapy.maps.WcsNDMap.binary_erode` and
# `~gammapy.maps.WcsNDMap.binary_dilate` methods, respectively.
#

fig, (ax1, ax2) = plt.subplots(
    figsize=(11, 5), ncols=2, subplot_kw={"projection": significance_mask_inv.geom.wcs}
)

mask = significance_mask_inv.binary_erode(width=0.2 * u.deg, kernel="disk")
mask.plot(ax=ax1)

mask = significance_mask_inv.binary_dilate(width=0.2 * u.deg)
mask.plot(ax=ax2)
plt.show()


######################################################################
# Boundary mask
# ~~~~~~~~~~~~~
#
# In the following example we use the Fermi dataset previously loaded and
# add its ``mask_fit`` taking into account a margin based on the psf
# width. The margin width is determined using the `~gammapy.irf.PSFMap.containment_radius`
# method of the psf object and the mask is created using the
# `~gammapy.maps.WcsGeom.boundary_mask` method available on the geometry object.
#

# get PSF 95% containment radius
energy_true = dataset.exposure.geom.axes[0].center
psf_r95 = dataset.psf.containment_radius(fraction=0.95, energy_true=energy_true)
plt.show()

# create mask_fit with margin based on PSF
mask_fit = dataset.counts.geom.boundary_mask(psf_r95.max())
dataset.mask_fit = mask_fit
dataset.mask_fit.sum_over_axes().plot()
plt.show()

######################################################################
# Reading and writing masks
# -------------------------
#
# `gammapy.maps` can directly read/write maps with boolean content as
# follows:
#

# To save masks to disk
mask_map.write("exclusion_mask.fits", overwrite="True")

# To read maps from disk
mask_map = Map.read("exclusion_mask.fits")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
import scipy.special
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import SkyCoord
from gammapy.irf import IRF
from gammapy.makers.utils import make_edisp_kernel_map, make_map_exposure_true_energy
from gammapy.maps import MapAxes, MapAxis, WcsGeom


class EffectiveArea3D(IRF):
    tag = "aeff_3d"
    required_axes = ["energy_true", "fov_lon", "fov_lat"]
    default_unit = u.m**2


@pytest.fixture
def aeff_3d():
    energy_axis = MapAxis.from_energy_edges(
        [0.1, 0.3, 1.0, 3.0, 10.0] * u.TeV, name="energy_true"
    )
    fov_lon_axis = MapAxis.from_edges([-1.5, -0.5, 0.5, 1.5] * u.deg, name="fov_lon")
    fov_lat_axis = MapAxis.from_edges([-1.5, -0.5, 0.5, 1.5] * u.deg, name="fov_lat")

    data = np.ones((4, 3, 3))
    for i in range(1, 4):
        data[i] = data[i - 1] * 1.5

    return EffectiveArea3D(
        [energy_axis, fov_lon_axis, fov_lat_axis], data=data, unit=u.m**2
    )


class EnergyDispersion3D(IRF):
    tag = "edisp_3d"
    required_axes = ["energy_true", "migra", "fov_lon", "fov_lat"]
    default_unit = u.one

    @classmethod
    def from_gauss(
        cls,
        energy_axis_true,
        migra_axis,
        fov_lon_axis,
        fov_lat_axis,
        bias,
        sigma,
        pdf_threshold=1e-6,
    ):
        axes = MapAxes([energy_axis_true, migra_axis, fov_lon_axis, fov_lat_axis])
        coords = axes.get_coord(mode="edges", axis_name="migra")

        migra_min = coords["migra"][:, :-1, :]
        migra_max = coords["migra"][:, 1:, :]

        # Analytical formula for integral of Gaussian
        s = np.sqrt(2) * sigma
        t1 = (migra_max - 1 - bias) / s
        t2 = (migra_min - 1 - bias) / s
        pdf = (scipy.special.erf(t1) - scipy.special.erf(t2)) / 2
        pdf = pdf / (migra_max - migra_min)

        r1 = np.rollaxis(pdf, -1, 1)
        r2 = np.rollaxis(r1, 0, -1)
        data = r2 * np.ones(axes.shape)

        data[data < pdf_threshold] = 0

        return cls(
            axes=axes,
            data=data.value,
        )


@pytest.fixture
def edisp_3d():
    energy_axis_true = MapAxis.from_energy_bounds(
        "0.1 TeV", "100 TeV", nbin=5, name="energy_true"
    )

    migra_axis = MapAxis.from_bounds(0, 4, nbin=10, node_type="edges", name="migra")

    fov_lon_axis = MapAxis.from_edges([-1.5, -0.5, 0.5, 1.5] * u.deg, name="fov_lon")
    fov_lat_axis = MapAxis.from_edges([-1.5, -0.5, 0.5, 1.5] * u.deg, name="fov_lat")

    energy_true = energy_axis_true.edges[:-1]
    sigma = 0.15 / (energy_true / (1 * u.TeV)).value ** 0.3
    bias = 1e-3 * (energy_true - 1 * u.TeV).value

    return EnergyDispersion3D.from_gauss(
        energy_axis_true=energy_axis_true,
        migra_axis=migra_axis,
        fov_lon_axis=fov_lon_axis,
        fov_lat_axis=fov_lat_axis,
        bias=bias,
        sigma=sigma,
    )


def test_aeff_3d(aeff_3d):
    res = aeff_3d.evaluate(
        fov_lon=[-0.5, 0.8] * u.deg,
        fov_lat=[-0.5, 1.0] * u.deg,
        energy_true=[0.2, 8.0] * u.TeV,
    )
    assert_allclose(res.data, [1.06246937, 3.74519106], rtol=1e-5)

    axis = MapAxis.from_energy_bounds(0.1 * u.TeV, 10 * u.TeV, 6, name="energy_true")
    pointing = SkyCoord(2, 1, unit="deg")
    geom = WcsGeom.create(npix=(4, 3), binsz=2, axes=[axis], skydir=pointing)

    exposure_map = make_map_exposure_true_energy(
        pointing=pointing, livetime="42 h", aeff=aeff_3d, geom=geom
    )
    assert_allclose(
        exposure_map.data[3][1][1:3], [323894.44971479, 323894.44971479], rtol=1e-5
    )


def test_edisp_3d(edisp_3d):
    energy = [1, 2] * u.TeV
    migra = np.array([0.98, 0.97, 0.7])
    fov_lon = [0.1, 1.5] * u.deg
    fov_lat = [0.0, 0.3] * u.deg

    eval = edisp_3d.evaluate(
        energy_true=energy.reshape(-1, 1, 1, 1),
        migra=migra.reshape(1, -1, 1, 1),
        fov_lon=fov_lon.reshape(1, 1, -1, 1),
        fov_lat=fov_lat.reshape(1, 1, 1, -1),
    )

    assert_allclose(
        eval[0][2], [[0.71181427, 0.71181427], [0.71181427, 0.71181427]], rtol=1e-3
    )
    etrue = MapAxis.from_energy_bounds(0.5, 2, 6, unit="TeV", name="energy_true")
    ereco = MapAxis.from_energy_bounds(0.5, 2, 3, unit="TeV", name="energy")
    pointing = SkyCoord(2, 1, unit="deg")
    geom = WcsGeom.create(10, binsz=0.5, axes=[ereco, etrue], skydir=pointing)

    edispmap = make_edisp_kernel_map(edisp_3d, pointing, geom)
    assert_allclose(edispmap.edisp_map.data[3][1][5][5], 0.623001, rtol=1e-3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import os
import yaml
from gammapy.scripts.download import RELEASE, cli_download_datasets
from gammapy.scripts.info import cli_info

log = logging.getLogger(__name__)


def check_tutorials_setup(download_datasets_path="./gammapy-data"):
    """Check tutorials setup and download data if not available.

    Parameters
    ----------
    download_datasets_path : str, optional
        Path to download the data. Default is "./gammapy-data".
    """
    if "GAMMAPY_DATA" not in os.environ:
        log.info(
            "Missing example datasets, downloading to {download_datasets_path} now..."
        )
        cli_download_datasets.callback(out=download_datasets_path, release=RELEASE)
        os.environ["GAMMAPY_DATA"] = download_datasets_path

    cli_info.callback(system=True, version=True, dependencies=True, envvar=True)


def yaml_checksum(yaml_content):
    """Compute a MD5 checksum for a given yaml string input."""
    import hashlib

    # Calculate the MD5 checksum
    checksum = hashlib.md5(yaml_content.encode("utf-8")).hexdigest()

    return checksum


def verify_checksum(yaml_content, checksum):
    """Compare MD5 checksum for yaml_content with input checksum."""
    return yaml_checksum(yaml_content) == checksum


def add_checksum(yaml_str, sort_keys=False, indent=4, width=50):
    """Append a checksum at the end of the yaml string."""
    checksum = {"checksum": yaml_checksum(yaml_str)}
    checksum = yaml.dump(
        checksum,
        sort_keys=True,
        indent=4,  # indent,
        width=80,  # width,
        default_flow_style=False,
    )
    return yaml_str + checksum



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose, assert_equal
import astropy.units as u
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
from astropy.utils.exceptions import AstropyUserWarning
from gammapy.data import GTI
from gammapy.datasets import Datasets, SpectrumDataset, SpectrumDatasetOnOff
from gammapy.irf import EDispKernelMap, EffectiveAreaTable2D
from gammapy.makers.utils import make_map_exposure_true_energy
from gammapy.maps import LabelMapAxis, MapAxis, RegionGeom, RegionNDMap, WcsGeom
from gammapy.modeling import Fit
from gammapy.modeling.models import (
    ConstantSpectralModel,
    ExpCutoffPowerLawSpectralModel,
    Models,
    PowerLawSpectralModel,
    SkyModel,
    GaussianPrior,
)
from gammapy.utils.random import get_random_state
from gammapy.utils.regions import compound_region_to_regions
from gammapy.utils.testing import assert_time_allclose, mpl_plot_check, requires_data


def test_data_shape(spectrum_dataset):
    assert spectrum_dataset.data_shape[0] == 30


def test_str(spectrum_dataset):
    assert "SpectrumDataset" in str(spectrum_dataset)


def test_energy_range(spectrum_dataset):
    e_min, e_max = spectrum_dataset.energy_range
    assert e_min.unit == u.TeV
    assert e_max.unit == u.TeV
    assert_allclose(e_min, 0.1)
    assert_allclose(e_max, 10.0)


def test_info_dict(spectrum_dataset):
    info_dict = spectrum_dataset.info_dict()

    assert_allclose(info_dict["counts"], 907010)
    assert_allclose(info_dict["background"], 3000.0)

    assert_allclose(info_dict["sqrt_ts"], 2924.522174)
    assert_allclose(info_dict["excess"], 904010)
    assert_allclose(info_dict["ontime"].value, 216000)

    assert info_dict["name"] == "test"


def test_set_model(spectrum_dataset):
    spectrum_dataset = spectrum_dataset.copy()
    spectral_model = PowerLawSpectralModel()
    model = SkyModel(spectral_model=spectral_model, name="test")
    spectrum_dataset.models = model
    assert spectrum_dataset.models["test"] is model

    models = Models([model])
    spectrum_dataset.models = models
    assert spectrum_dataset.models["test"] is model


def test_spectrum_dataset_fits_io(spectrum_dataset, tmp_path):
    spectrum_dataset.meta_table = Table(
        data=[[1.0 * u.h], [111]], names=["livetime", "obs_id"]
    )
    hdulist = spectrum_dataset.to_hdulist()
    actual = [hdu.name for hdu in hdulist]
    desired = [
        "PRIMARY",
        "COUNTS",
        "COUNTS_BANDS",
        "COUNTS_REGION",
        "EXPOSURE",
        "EXPOSURE_BANDS",
        "EXPOSURE_REGION",
        "BACKGROUND",
        "BACKGROUND_BANDS",
        "BACKGROUND_REGION",
        "GTI",
        "META_TABLE",
    ]

    assert actual == desired

    spectrum_dataset.write(tmp_path / "test.fits")
    dataset_new = SpectrumDataset.read(tmp_path / "test.fits", name="test")

    assert_allclose(spectrum_dataset.counts.data, dataset_new.counts.data)
    assert_allclose(
        spectrum_dataset.npred_background().data, dataset_new.npred_background().data
    )
    assert dataset_new.edisp is None
    assert dataset_new.edisp is None
    assert dataset_new.name == "test"

    assert_allclose(spectrum_dataset.exposure.data, dataset_new.exposure.data)
    assert spectrum_dataset.counts.geom == dataset_new.counts.geom

    assert_allclose(dataset_new.meta_table["obs_id"], 111)


def test_npred_models():
    e_reco = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=3)

    geom = RegionGeom(region=None, axes=[e_reco])

    spectrum_dataset = SpectrumDataset.create(geom=geom)
    spectrum_dataset.exposure.quantity = 1e10 * u.Unit("cm2 h")

    pwl_1 = PowerLawSpectralModel(index=2)
    pwl_2 = PowerLawSpectralModel(index=2)
    model_1 = SkyModel(spectral_model=pwl_1)
    model_2 = SkyModel(spectral_model=pwl_2)

    spectrum_dataset.models = Models([model_1, model_2])
    npred = spectrum_dataset.npred()

    assert_allclose(npred.data.sum(), 64.8)

    npred_sig = spectrum_dataset.npred_signal()
    assert_allclose(npred_sig.data.sum(), 64.8)

    npred_sig_model1 = spectrum_dataset.npred_signal(model_names=[model_1.name])
    assert_allclose(npred_sig_model1.data.sum(), 32.4)

    assert_allclose(
        spectrum_dataset.npred_signal(
            model_names=[model_1.name, model_2.name]
        ).data.sum(),
        64.8,
    )

    npred_model1_not_stack = spectrum_dataset.npred_signal(
        model_names=[model_1.name], stack=False
    )
    assert_allclose(npred_model1_not_stack.geom.data_shape, (1, 3, 1, 1))
    assert_allclose(npred_model1_not_stack.data.sum(), 32.4)
    assert isinstance(npred_model1_not_stack.geom.axes[-1], LabelMapAxis)
    assert npred_model1_not_stack.geom.axes[-1].name == "models"
    assert_equal(npred_model1_not_stack.geom.axes[-1].center, [model_1.name])

    npred_all_models_not_stack = spectrum_dataset.npred_signal(
        model_names=[model_1.name, model_2.name], stack=False
    )
    assert_allclose(npred_all_models_not_stack.geom.data_shape, (2, 3, 1, 1))
    assert_allclose(
        npred_all_models_not_stack.sum_over_axes(["models"]).data.sum(), 64.8
    )


def test_npred_spatial_model(spectrum_dataset):
    model = SkyModel.create("pl", "gauss", name="test")

    spectrum_dataset.models = [model]

    npred = spectrum_dataset.npred()
    model.spatial_model.sigma.value = 1.0
    npred_large_sigma = spectrum_dataset.npred()

    assert_allclose(npred.data.sum(), 3000)
    assert_allclose(npred_large_sigma.data.sum(), 3000)
    assert spectrum_dataset.evaluators["test"].psf is None


def test_fit(spectrum_dataset):
    """Simple CASH fit to the on vector"""
    fit = Fit()
    result = fit.run(datasets=[spectrum_dataset])
    assert result.success
    assert "minuit" in str(result)

    npred = spectrum_dataset.npred().data.sum()
    assert_allclose(npred, 907012.186399, rtol=1e-3)
    assert_allclose(result.total_stat, -18087404.624, rtol=1e-3)

    pars = spectrum_dataset.models.parameters
    assert_allclose(pars["index"].value, 2.1, rtol=1e-2)
    assert_allclose(pars["index"].error, 0.001276, rtol=1e-2)

    assert_allclose(pars["amplitude"].value, 1e5, rtol=1e-3)
    assert_allclose(pars["amplitude"].error, 153.450825, rtol=1e-2)


def test_spectrum_dataset_create():
    e_reco = MapAxis.from_edges(u.Quantity([0.1, 1, 10.0], "TeV"), name="energy")
    e_true = MapAxis.from_edges(
        u.Quantity([0.05, 0.5, 5, 20.0], "TeV"), name="energy_true"
    )
    geom = RegionGeom(region=None, axes=[e_reco])
    empty_spectrum_dataset = SpectrumDataset.create(
        geom, energy_axis_true=e_true, name="test"
    )

    assert empty_spectrum_dataset.name == "test"
    assert empty_spectrum_dataset.counts.data.sum() == 0
    assert empty_spectrum_dataset.data_shape[0] == 2
    assert empty_spectrum_dataset.background.data.sum() == 0
    assert empty_spectrum_dataset.background.geom.axes[0].nbin == 2
    assert empty_spectrum_dataset.exposure.geom.axes[0].nbin == 3
    assert empty_spectrum_dataset.edisp.edisp_map.geom.axes["energy"].nbin == 2
    assert empty_spectrum_dataset.gti.time_sum.value == 0
    assert len(empty_spectrum_dataset.gti.table) == 0
    assert np.isnan(empty_spectrum_dataset.energy_range[0])
    assert_allclose(empty_spectrum_dataset.mask_safe, 0)


def test_spectrum_dataset_stack_diagonal_safe_mask(spectrum_dataset):
    geom = spectrum_dataset.counts.geom

    energy = MapAxis.from_energy_bounds("0.1 TeV", "10 TeV", nbin=30)
    energy_true = MapAxis.from_energy_bounds(
        "0.1 TeV", "10 TeV", nbin=30, name="energy_true"
    )

    aeff = EffectiveAreaTable2D.from_parametrization(
        energy_axis_true=energy_true, instrument="HESS"
    )

    livetime = 100 * u.s
    gti = GTI.create(start=0 * u.s, stop=livetime)

    geom_true = geom.as_energy_true
    exposure = make_map_exposure_true_energy(
        geom=geom_true, livetime=livetime, pointing=geom_true.center_skydir, aeff=aeff
    )

    edisp = EDispKernelMap.from_diagonal_response(
        energy, energy_true, geom=geom.to_image()
    )
    edisp.exposure_map.data = exposure.data[:, :, np.newaxis, :]

    background = spectrum_dataset.background

    mask_safe = RegionNDMap.from_geom(geom=geom, dtype=bool)
    mask_safe.data += True

    spectrum_dataset1 = SpectrumDataset(
        name="ds1",
        counts=spectrum_dataset.counts.copy(),
        exposure=exposure.copy(),
        edisp=edisp.copy(),
        background=background.copy(),
        gti=gti.copy(),
        mask_safe=mask_safe,
    )

    livetime2 = 0.5 * livetime
    gti2 = GTI.create(start=200 * u.s, stop=200 * u.s + livetime2)
    bkg2 = RegionNDMap.from_geom(geom=geom, data=2 * background.data)

    geom = spectrum_dataset.counts.geom
    data = np.ones(spectrum_dataset.data_shape, dtype="bool")
    data[0] = False
    safe_mask2 = RegionNDMap.from_geom(geom=geom, data=data)
    exposure2 = exposure.copy()

    edisp = edisp.copy()
    edisp.exposure_map.data = exposure2.data[:, :, np.newaxis, :]
    spectrum_dataset2 = SpectrumDataset(
        name="ds2",
        counts=spectrum_dataset.counts.copy(),
        exposure=exposure2,
        edisp=edisp,
        background=bkg2,
        mask_safe=safe_mask2,
        gti=gti2,
    )

    spectrum_dataset1.stack(spectrum_dataset2)

    reference = spectrum_dataset.counts.data
    assert_allclose(spectrum_dataset1.counts.data[1:], reference[1:] * 2)
    assert_allclose(spectrum_dataset1.counts.data[0], 141363)
    assert_allclose(
        spectrum_dataset1.exposure.quantity[0], 4.755644e09 * u.Unit("cm2 s")
    )
    assert_allclose(spectrum_dataset1.background.data[1:], 3 * background.data[1:])
    assert_allclose(spectrum_dataset1.background.data[0], background.data[0])

    kernel = edisp.get_edisp_kernel()
    kernel_stacked = spectrum_dataset1.edisp.get_edisp_kernel()

    assert_allclose(kernel_stacked.pdf_matrix[1:], kernel.pdf_matrix[1:])
    assert_allclose(kernel_stacked.pdf_matrix[0], 0.5 * kernel.pdf_matrix[0])


def test_spectrum_dataset_stack_nondiagonal_no_bkg(spectrum_dataset):
    energy = spectrum_dataset.counts.geom.axes["energy"]
    geom = spectrum_dataset.counts.geom

    edisp1 = EDispKernelMap.from_gauss(
        energy_axis=energy,
        energy_axis_true=energy.copy(name="energy_true"),
        sigma=0.1,
        bias=0,
        geom=geom.to_image(),
    )
    edisp1.exposure_map.data += 1

    aeff = EffectiveAreaTable2D.from_parametrization(
        energy_axis_true=energy.copy(name="energy_true"), instrument="HESS"
    )

    livetime = 100 * u.s

    geom_true = geom.as_energy_true
    exposure = make_map_exposure_true_energy(
        geom=geom_true, livetime=livetime, pointing=geom_true.center_skydir, aeff=aeff
    )

    geom = spectrum_dataset.counts.geom
    counts = RegionNDMap.from_geom(geom=geom)

    gti = GTI.create(start=0 * u.s, stop=livetime)
    spectrum_dataset1 = SpectrumDataset(
        counts=counts,
        exposure=exposure,
        edisp=edisp1,
        meta_table=Table({"OBS_ID": [0]}),
        gti=gti.copy(),
    )

    edisp2 = EDispKernelMap.from_gauss(
        energy_axis=energy,
        energy_axis_true=energy.copy(name="energy_true"),
        sigma=0.2,
        bias=0.0,
        geom=geom,
    )
    edisp2.exposure_map.data += 1

    gti2 = GTI.create(start=100 * u.s, stop=200 * u.s)

    spectrum_dataset2 = SpectrumDataset(
        counts=counts,
        exposure=exposure.copy(),
        edisp=edisp2,
        meta_table=Table({"OBS_ID": [1]}),
        gti=gti2,
    )
    spectrum_dataset1.stack(spectrum_dataset2)

    assert_allclose(spectrum_dataset1.meta_table["OBS_ID"][0], [0, 1])

    assert spectrum_dataset1.background_model is None
    assert_allclose(spectrum_dataset1.gti.time_sum.to_value("s"), 200)
    assert_allclose(
        spectrum_dataset1.exposure.quantity[2].to_value("m2 s"), 1573851.079861
    )
    kernel = edisp1.get_edisp_kernel()
    assert_allclose(kernel.get_bias(1 * u.TeV), 0.0, atol=1.2e-3)
    assert_allclose(kernel.get_resolution(1 * u.TeV), 0.1581, atol=1e-2)


def test_peek(spectrum_dataset):
    with mpl_plot_check():
        spectrum_dataset.peek()

    with mpl_plot_check():
        spectrum_dataset.plot_fit()

    spectrum_dataset.edisp = None
    with mpl_plot_check():
        spectrum_dataset.peek()


class TestSpectrumOnOff:
    """Test ON OFF SpectrumDataset"""

    def setup_method(self):
        etrue = np.logspace(-1, 1, 10) * u.TeV
        self.e_true = MapAxis.from_energy_edges(etrue, name="energy_true")
        ereco = np.logspace(-1, 1, 5) * u.TeV
        elo = ereco[:-1]
        self.e_reco = MapAxis.from_energy_edges(ereco, name="energy")

        start = u.Quantity([0], "s")
        stop = u.Quantity([1000], "s")
        time_ref = Time("2010-01-01 00:00:00.0")
        self.gti = GTI.create(start, stop, time_ref)
        self.livetime = self.gti.time_sum

        self.wcs = WcsGeom.create(npix=300, binsz=0.01, frame="icrs").wcs

        self.aeff = RegionNDMap.create(
            region="icrs;circle(0.,1.,0.1)",
            wcs=self.wcs,
            axes=[self.e_true],
            unit="cm2",
        )
        self.aeff.data += 1

        data = np.ones(elo.shape)
        data[-1] = 0  # to test stats calculation with empty bins

        axis = MapAxis.from_edges(ereco, name="energy", interp="log")
        self.on_counts = RegionNDMap.create(
            region="icrs;circle(0.,1.,0.1)",
            wcs=self.wcs,
            axes=[axis],
            meta={"EXPOSURE": self.livetime.to_value("s")},
        )
        self.on_counts.data += 1
        self.on_counts.data[-1] = 0

        self.off_counts = RegionNDMap.create(
            region="icrs;box(0.,1.,0.1, 0.2,30);box(-1.,-1.,0.1, 0.2,150)",
            wcs=self.wcs,
            axes=[axis],
        )
        self.off_counts.data += 10

        acceptance = RegionNDMap.from_geom(self.on_counts.geom)
        acceptance.data += 1

        data = np.ones(elo.shape)
        data[-1] = 0

        acceptance_off = RegionNDMap.from_geom(self.off_counts.geom)
        acceptance_off.data += 10

        self.edisp = EDispKernelMap.from_diagonal_response(
            self.e_reco, self.e_true, self.on_counts.geom.to_image()
        )

        exposure = self.aeff * self.livetime
        exposure.meta["livetime"] = self.livetime

        mask_safe = RegionNDMap.from_geom(self.on_counts.geom, dtype=bool)
        mask_safe.data += True

        self.dataset = SpectrumDatasetOnOff(
            counts=self.on_counts,
            counts_off=self.off_counts,
            exposure=exposure,
            edisp=self.edisp,
            acceptance=acceptance,
            acceptance_off=acceptance_off,
            name="test",
            gti=self.gti,
            mask_safe=mask_safe,
        )

    def test_spectrum_dataset_on_off_create(self):
        e_reco = MapAxis.from_edges(u.Quantity([0.1, 1, 10.0], "TeV"), name="energy")
        e_true = MapAxis.from_edges(
            u.Quantity([0.05, 0.5, 5, 20.0], "TeV"), name="energy_true"
        )
        geom = RegionGeom(region=None, axes=[e_reco])
        empty_dataset = SpectrumDatasetOnOff.create(geom=geom, energy_axis_true=e_true)

        assert empty_dataset.counts.data.sum() == 0
        assert empty_dataset.data_shape[0] == 2
        assert empty_dataset.counts_off.data.sum() == 0
        assert empty_dataset.counts_off.geom.axes[0].nbin == 2
        assert_allclose(empty_dataset.acceptance_off, 0)
        assert_allclose(empty_dataset.acceptance, 0)
        assert empty_dataset.acceptance.data.shape[0] == 2
        assert empty_dataset.acceptance_off.data.shape[0] == 2
        assert empty_dataset.gti.time_sum.value == 0
        assert len(empty_dataset.gti.table) == 0
        assert np.isnan(empty_dataset.energy_range[0])

    def test_create_stack(self):
        geom = RegionGeom(region=None, axes=[self.e_reco])

        stacked = SpectrumDatasetOnOff.create(geom=geom, energy_axis_true=self.e_true)
        stacked.mask_safe.data += True

        stacked.stack(self.dataset)
        e_min_stacked, e_max_stacked = stacked.energy_range
        e_min_dataset, e_max_dataset = self.dataset.energy_range
        assert_allclose(e_min_stacked, e_min_dataset)
        assert_allclose(e_max_stacked, e_max_dataset)

    def test_alpha(self):
        assert self.dataset.alpha.data.shape == (4, 1, 1)
        assert_allclose(self.dataset.alpha.data, 0.1)

    def test_npred_no_edisp(self):
        const = 1 * u.Unit("cm-2 s-1 TeV-1")
        model = SkyModel(spectral_model=ConstantSpectralModel(const=const))
        livetime = 1 * u.s

        aeff = RegionNDMap.create(
            region=self.aeff.geom.region,
            unit="cm2",
            axes=[self.e_reco.copy(name="energy_true")],
        )

        aeff.data += 1
        dataset = SpectrumDatasetOnOff(
            counts=self.on_counts,
            counts_off=self.off_counts,
            exposure=aeff * livetime,
            models=model,
        )
        energy = aeff.geom.axes[0].edges
        expected = aeff.data[0] * (energy[-1] - energy[0]) * const * livetime

        assert_allclose(dataset.npred_signal().data.sum(), expected.value)

    def test_to_spectrum_dataset(self):
        ds = self.dataset.to_spectrum_dataset()

        assert isinstance(ds, SpectrumDataset)
        assert_allclose(ds.background.data.sum(), 4)

    def test_peek(self):
        dataset = self.dataset.copy()
        dataset.models = SkyModel(spectral_model=PowerLawSpectralModel())

        with mpl_plot_check():
            dataset.peek()

    def test_plot_fit(self):
        dataset = self.dataset.copy()
        dataset.models = SkyModel(spectral_model=PowerLawSpectralModel())

        with mpl_plot_check():
            dataset.plot_fit()

    def test_to_from_ogip_files(self, tmp_path):
        dataset = self.dataset.copy(name="test")
        dataset.write(tmp_path / "test.fits")
        newdataset = SpectrumDatasetOnOff.read(tmp_path / "test.fits")

        expected_regions = compound_region_to_regions(self.off_counts.geom.region)
        regions = compound_region_to_regions(newdataset.counts_off.geom.region)

        assert newdataset.counts.meta["RESPFILE"] == "test_rmf.fits"
        assert newdataset.counts.meta["BACKFILE"] == "test_bkg.fits"
        assert newdataset.counts.meta["ANCRFILE"] == "test_arf.fits"

        assert_allclose(self.on_counts.data, newdataset.counts.data)
        assert_allclose(self.off_counts.data, newdataset.counts_off.data)
        assert_allclose(self.edisp.edisp_map.data, newdataset.edisp.edisp_map.data)
        assert_time_allclose(newdataset.gti.time_start, dataset.gti.time_start)

        assert len(regions) == len(expected_regions)
        assert regions[0].center.is_equivalent_frame(expected_regions[0].center)
        assert_allclose(regions[1].angle, expected_regions[1].angle)

    def test_from_ogip_files_overwrite_name(self, tmp_path):
        dataset = self.dataset.copy(name="test")
        dataset.write(tmp_path / "test.fits")
        new_dataset = SpectrumDatasetOnOff.read(tmp_path / "test.fits")
        assert new_dataset.name == dataset.name

        new_dataset = SpectrumDatasetOnOff.read(tmp_path / "test.fits", name="new_name")
        assert new_dataset.name == "new_name"

    def test_to_from_ogip_files_no_mask(self, tmp_path):
        dataset = self.dataset.copy(name="test")
        dataset.mask_safe = None
        dataset.write(tmp_path / "test.fits")
        newdataset = SpectrumDatasetOnOff.read(tmp_path / "test.fits")

        assert_allclose(newdataset.mask_safe.data, True)

    def test_to_from_ogip_files_zip(self, tmp_path):
        dataset = self.dataset.copy(name="test")
        dataset.write(tmp_path / "test.fits.gz")
        newdataset = SpectrumDatasetOnOff.read(tmp_path / "test.fits.gz")

        assert newdataset.counts.meta["RESPFILE"] == "test_rmf.fits.gz"
        assert newdataset.counts.meta["BACKFILE"] == "test_bkg.fits.gz"
        assert newdataset.counts.meta["ANCRFILE"] == "test_arf.fits.gz"

    def test_to_from_ogip_files_no_edisp(self, tmp_path):
        mask_safe = RegionNDMap.from_geom(self.on_counts.geom, dtype=bool)
        mask_safe.data += True

        acceptance = RegionNDMap.from_geom(self.on_counts.geom, data=1.0)

        exposure = self.aeff * self.livetime
        exposure.meta["livetime"] = self.livetime

        dataset = SpectrumDatasetOnOff(
            counts=self.on_counts,
            exposure=exposure,
            mask_safe=mask_safe,
            acceptance=acceptance,
            name="test",
        )
        dataset.write(tmp_path / "pha_obstest.fits")
        newdataset = SpectrumDatasetOnOff.read(tmp_path / "pha_obstest.fits")

        assert_allclose(self.on_counts.data, newdataset.counts.data)
        assert newdataset.counts_off is None
        assert newdataset.edisp is None
        assert newdataset.gti is None

    def test_to_ogip_files_checksum(self, tmp_path):
        dataset = self.dataset.copy(name="test")
        dataset.write(tmp_path / "test.fits", format="ogip", checksum=True)

        for name in ["test.fits", "test_arf.fits", "test_rmf.fits", "test_bkg.fits"]:
            # TODO: this should not emit AstropyUserWarning
            with fits.open(tmp_path / name, checksum=True) as hdul:
                for hdu in hdul:
                    assert "CHECKSUM" in hdu.header
                    assert "DATASUM" in hdu.header

        def replace_in_fits_header(filename, string):
            with open(filename, "r+b") as file:
                chunk = file.read(10000)
                index = chunk.find(string.encode("ascii"))
                file.seek(index)
                file.write("bad".encode("ascii"))

        path = tmp_path / "test.fits"
        replace_in_fits_header(path, "unknown")
        with pytest.warns(AstropyUserWarning):
            SpectrumDatasetOnOff.read(path, checksum=True)

    def test_spectrum_dataset_onoff_fits_io(self, tmp_path):
        self.dataset.write(tmp_path / "test.fits", format="gadf")
        d1 = SpectrumDatasetOnOff.read(tmp_path / "test.fits", format="gadf")
        assert isinstance(d1.counts.geom, RegionGeom)
        assert d1.exposure == self.dataset.exposure
        assert_allclose(d1.counts_off.data, self.dataset.counts_off.data)

    def test_energy_mask(self):
        mask = self.dataset.counts.geom.energy_mask(
            energy_min=0.3 * u.TeV, energy_max=6 * u.TeV
        )
        desired = [False, True, True, False]
        assert_allclose(mask.data[:, 0, 0], desired)

        mask = self.dataset.counts.geom.energy_mask(energy_max=6 * u.TeV)
        desired = [True, True, True, False]
        assert_allclose(mask.data[:, 0, 0], desired)

        mask = self.dataset.counts.geom.energy_mask(energy_min=1 * u.TeV)
        desired = [False, False, True, True]
        assert_allclose(mask.data[:, 0, 0], desired)

    def test_str(self):
        model = SkyModel(spectral_model=PowerLawSpectralModel())
        dataset = SpectrumDatasetOnOff(
            counts=self.on_counts,
            counts_off=self.off_counts,
            models=model,
            exposure=self.aeff * self.livetime,
            edisp=self.edisp,
            acceptance=RegionNDMap.from_geom(geom=self.on_counts.geom, data=1),
            acceptance_off=RegionNDMap.from_geom(geom=self.off_counts.geom, data=10),
        )
        assert "SpectrumDatasetOnOff" in str(dataset)
        assert "wstat" in str(dataset)

    def test_fake(self):
        """Test the fake dataset"""
        source_model = SkyModel(spectral_model=PowerLawSpectralModel())
        dataset = SpectrumDatasetOnOff(
            name="test",
            counts=self.on_counts,
            counts_off=self.off_counts,
            models=source_model,
            exposure=self.aeff * self.livetime,
            edisp=self.edisp,
            acceptance=RegionNDMap.from_geom(geom=self.on_counts.geom, data=1),
            acceptance_off=RegionNDMap.from_geom(geom=self.off_counts.geom, data=10),
        )
        real_dataset = dataset.copy()

        background = RegionNDMap.from_geom(dataset.counts.geom)
        background.data += 1
        dataset.fake(npred_background=background, random_state=314)

        assert real_dataset.counts.data.shape == dataset.counts.data.shape
        assert real_dataset.counts_off.data.shape == dataset.counts_off.data.shape
        assert dataset.counts_off.data.sum() == 39
        assert dataset.counts.data.sum() == 5

    def test_info_dict(self):
        info_dict = self.dataset.info_dict()

        assert_allclose(info_dict["counts"], 3)
        assert_allclose(info_dict["counts_off"], 40)
        assert_allclose(info_dict["acceptance"], 4)
        assert_allclose(info_dict["acceptance_off"], 40)

        assert_allclose(info_dict["alpha"], 0.1)
        assert_allclose(info_dict["excess"], -1, rtol=1e-2)
        assert_allclose(info_dict["ontime"].value, 1e3)
        assert_allclose(info_dict["sqrt_ts"], -0.501005, rtol=1e-2)

        assert info_dict["name"] == "test"

    def test_resample_energy_axis(self):
        axis = MapAxis.from_edges([0.1, 1, 10] * u.TeV, name="energy", interp="log")
        grouped = self.dataset.resample_energy_axis(energy_axis=axis)

        assert grouped.counts.data.shape == (2, 1, 1)
        # exposure should be untouched
        assert_allclose(grouped.exposure.data, 1000)
        assert_allclose(np.squeeze(grouped.counts), [2, 1])
        assert_allclose(np.squeeze(grouped.counts_off), [20, 20])
        assert grouped.edisp.edisp_map.data.shape == (9, 2, 1, 1)
        assert_allclose(np.squeeze(grouped.acceptance), [2, 2])
        assert_allclose(np.squeeze(grouped.acceptance_off), [20, 20])

    def test_to_image(self):
        grouped = self.dataset.to_image()

        assert grouped.counts.data.shape == (1, 1, 1)
        # exposure should be untouched
        assert_allclose(grouped.exposure.data, 1000)
        assert_allclose(np.squeeze(grouped.counts), 3)
        assert_allclose(np.squeeze(grouped.counts_off), 40)
        assert grouped.edisp.edisp_map.data.shape == (9, 1, 1, 1)
        assert_allclose(np.squeeze(grouped.acceptance), 4)
        assert_allclose(np.squeeze(grouped.acceptance_off), 40)

    def test_spectrum_dataset_ogip_creation_metadata(self, tmp_path):
        dataset = self.dataset.copy(name="test")
        dataset.write(tmp_path / "test.fits", format="ogip")

        for name in ["test.fits", "test_arf.fits", "test_bkg.fits", "test_rmf.fits"]:
            with fits.open(tmp_path / name) as hdul:
                for hdu in hdul[1:]:
                    assert "CREATOR" in hdu.header
                    assert "CREATED" in hdu.header


@requires_data()
class TestSpectralFit:
    """Test fit in astrophysical scenario"""

    def setup_method(self):
        path = "$GAMMAPY_DATA/joint-crab/spectra/hess/"
        self.datasets = Datasets(
            [
                SpectrumDatasetOnOff.read(path + "pha_obs23523.fits"),
                SpectrumDatasetOnOff.read(path + "pha_obs23592.fits"),
            ]
        )
        self.pwl = SkyModel(
            spectral_model=PowerLawSpectralModel(
                index=2, amplitude=1e-12 * u.Unit("cm-2 s-1 TeV-1"), reference=1 * u.TeV
            )
        )

        self.ecpl = SkyModel(
            spectral_model=ExpCutoffPowerLawSpectralModel(
                index=2,
                amplitude=1e-12 * u.Unit("cm-2 s-1 TeV-1"),
                reference=1 * u.TeV,
                lambda_=0.1 / u.TeV,
            )
        )

        # Example fit for one observation
        self.datasets[0].models = self.pwl
        self.fit = Fit()

    def set_model(self, model):
        for obs in self.datasets:
            obs.models = model

    def test_basic_results(self):
        self.set_model(self.pwl)
        result = self.fit.run([self.datasets[0]])
        pars = self.datasets.parameters

        assert self.pwl is self.datasets[0].models[0]

        assert_allclose(result.total_stat, 38.343, rtol=1e-3)
        assert_allclose(pars["index"].value, 2.817, rtol=1e-3)
        assert pars["amplitude"].unit == "cm-2 s-1 TeV-1"
        assert_allclose(pars["amplitude"].value, 5.142e-11, rtol=1e-3)
        assert_allclose(self.datasets[0].npred().data[60], 0.6102, rtol=1e-3)
        pars.to_table()

    def test_basic_errors(self):
        self.set_model(self.pwl)
        self.fit.run([self.datasets[0]])
        pars = self.pwl.parameters

        assert_allclose(pars["index"].error, 0.149633, rtol=1e-3)
        assert_allclose(pars["amplitude"].error, 6.423139e-12, rtol=1e-3)
        pars.to_table()

    def test_ecpl_fit(self):
        self.set_model(self.ecpl)
        fit = Fit()
        fit.run([self.datasets[0]])

        actual = self.datasets.parameters["lambda_"].quantity
        assert actual.unit == "TeV-1"
        assert_allclose(actual.value, 0.145215, rtol=1e-2)

    def test_joint_fit(self):
        self.set_model(self.pwl)
        fit = Fit()
        fit.run(self.datasets)
        actual = self.datasets.parameters["index"].value
        assert_allclose(actual, 2.7806, rtol=1e-3)

        actual = self.datasets.parameters["amplitude"].quantity
        assert actual.unit == "cm-2 s-1 TeV-1"
        assert_allclose(actual.value, 5.200e-11, rtol=1e-3)

    def test_stats(self):
        dataset = self.datasets[0].copy()
        dataset.models = self.pwl

        fit = Fit()
        result = fit.run(datasets=[dataset])
        stats = dataset.stat_array()
        actual = np.sum(stats[dataset.mask_safe])

        desired = result.total_stat
        assert_allclose(actual, desired)

    def test_fit_range(self):
        # Fit range not restricted fit range should be the thresholds
        obs = self.datasets[0]
        actual = obs.energy_range[0]

        assert actual.unit == "keV"
        assert_allclose(actual, 8.912509e08)

    def test_no_edisp(self):
        dataset = self.datasets[0].copy()

        dataset.edisp = None
        dataset.models = self.pwl

        fit = Fit()
        fit.run(datasets=[dataset])
        assert_allclose(self.pwl.spectral_model.index.value, 2.7961, atol=0.02)

    def test_stacked_fit(self):
        dataset = self.datasets[0].copy()
        dataset.stack(self.datasets[1])
        dataset.models = SkyModel(PowerLawSpectralModel())

        fit = Fit()
        fit.run(datasets=[dataset])
        pars = dataset.models.parameters

        assert_allclose(pars["index"].value, 2.7767, rtol=1e-3)
        assert u.Unit(pars["amplitude"].unit) == "cm-2 s-1 TeV-1"
        assert_allclose(pars["amplitude"].value, 5.191e-11, rtol=1e-3)


def _read_hess_obs():
    path = "$GAMMAPY_DATA/joint-crab/spectra/hess/"
    obs1 = SpectrumDatasetOnOff.read(path + "pha_obs23523.fits")
    obs2 = SpectrumDatasetOnOff.read(path + "pha_obs23592.fits")
    return [obs1, obs2]


def make_gti(times, time_ref="2010-01-01"):
    return GTI.create(times["START"], times["STOP"], time_ref)


@requires_data("gammapy-data")
def make_observation_list():
    """obs with dummy IRF"""
    nbin = 3
    energy = np.logspace(-1, 1, nbin + 1) * u.TeV
    livetime = 2 * u.h
    data_on = np.arange(nbin)
    dataoff_1 = np.ones(3)
    dataoff_2 = np.ones(3) * 3
    dataoff_1[1] = 0
    dataoff_2[1] = 0

    axis = MapAxis.from_edges(energy, name="energy", interp="log")
    axis_true = axis.copy(name="energy_true")

    geom = RegionGeom(region=None, axes=[axis])
    geom_true = RegionGeom(region=None, axes=[axis_true])

    on_vector = RegionNDMap.from_geom(geom=geom, data=data_on)
    off_vector1 = RegionNDMap.from_geom(geom=geom, data=dataoff_1)
    off_vector2 = RegionNDMap.from_geom(geom=geom, data=dataoff_2)
    mask_safe = RegionNDMap.from_geom(geom, dtype=bool)
    mask_safe.data += True

    acceptance = RegionNDMap.from_geom(geom=geom, data=1)
    acceptance_off_1 = RegionNDMap.from_geom(geom=geom, data=2)
    acceptance_off_2 = RegionNDMap.from_geom(geom=geom, data=4)

    aeff = RegionNDMap.from_geom(geom_true, data=1, unit="m2")
    edisp = EDispKernelMap.from_gauss(
        energy_axis=axis, energy_axis_true=axis_true, sigma=0.2, bias=0, geom=geom
    )

    time_ref = Time("2010-01-01")
    gti1 = make_gti(
        {"START": [5, 6, 1, 2] * u.s, "STOP": [8, 7, 3, 4] * u.s}, time_ref=time_ref
    )
    gti2 = make_gti({"START": [14] * u.s, "STOP": [15] * u.s}, time_ref=time_ref)

    exposure = aeff * livetime
    exposure.meta["livetime"] = livetime

    obs1 = SpectrumDatasetOnOff(
        counts=on_vector,
        counts_off=off_vector1,
        exposure=exposure,
        edisp=edisp,
        mask_safe=mask_safe,
        acceptance=acceptance.copy(),
        acceptance_off=acceptance_off_1,
        name="1",
        gti=gti1,
    )
    obs2 = SpectrumDatasetOnOff(
        counts=on_vector,
        counts_off=off_vector2,
        exposure=exposure.copy(),
        edisp=edisp,
        mask_safe=mask_safe,
        acceptance=acceptance.copy(),
        acceptance_off=acceptance_off_2,
        name="2",
        gti=gti2,
    )

    obs_list = [obs1, obs2]
    return obs_list


@requires_data("gammapy-data")
class TestSpectrumDatasetOnOffStack:
    def setup_method(self):
        self.datasets = _read_hess_obs()
        # Change threshold to make stuff more interesting

        geom = self.datasets[0]._geom
        self.datasets[0].mask_safe = geom.energy_mask(
            energy_min=1.2 * u.TeV, energy_max=50 * u.TeV
        )

        mask = geom.energy_mask(energy_max=20 * u.TeV)
        self.datasets[1].mask_safe &= mask

        self.stacked_dataset = self.datasets[0].to_masked()
        self.stacked_dataset.stack(self.datasets[1])

    def test_basic(self):
        obs_1, obs_2 = self.datasets

        counts1 = obs_1.counts.data[obs_1.mask_safe].sum()
        counts2 = obs_2.counts.data[obs_2.mask_safe].sum()
        summed_counts = counts1 + counts2

        stacked_counts = self.stacked_dataset.counts.data.sum()

        off1 = obs_1.counts_off.data[obs_1.mask_safe].sum()
        off2 = obs_2.counts_off.data[obs_2.mask_safe].sum()
        summed_off = off1 + off2
        stacked_off = self.stacked_dataset.counts_off.data.sum()

        assert summed_counts == stacked_counts
        assert summed_off == stacked_off

    def test_thresholds(self):
        energy_min, energy_max = self.stacked_dataset.energy_range

        assert energy_min.unit == "keV"
        assert_allclose(energy_min, 8.912509e08, rtol=1e-3)

        assert energy_max.unit == "keV"
        assert_allclose(energy_max, 4.466836e10, rtol=1e-3)

    def test_verify_npred(self):
        """Verifying npred is preserved during the stacking"""
        pwl = SkyModel(
            spectral_model=PowerLawSpectralModel(
                index=2, amplitude=2e-11 * u.Unit("cm-2 s-1 TeV-1"), reference=1 * u.TeV
            )
        )

        self.stacked_dataset.models = pwl

        npred_stacked = self.stacked_dataset.npred_signal().data
        npred_stacked[~self.stacked_dataset.mask_safe.data] = 0
        npred_summed = np.zeros_like(npred_stacked)

        for dataset in self.datasets:
            dataset.models = pwl
            npred_summed[dataset.mask_safe] += dataset.npred_signal().data[
                dataset.mask_safe
            ]

        assert_allclose(npred_stacked, npred_summed, rtol=1e-6)

    def test_stack_backscal(self):
        """Verify backscal stacking"""
        obs1, obs2 = make_observation_list()
        obs1.stack(obs2)
        assert_allclose(obs1.alpha.data[0], 1.25 / 4.0)
        # When the OFF stack observation counts=0, the alpha is averaged on the
        # total OFF counts for each run.
        assert_allclose(obs1.alpha.data[1], 2.5 / 8.0)

    def test_stack_gti(self):
        obs1, obs2 = make_observation_list()
        obs1.stack(obs2)

        assert_allclose(obs1.gti.met_start.value, [1.0, 5.0, 14.0])
        assert_allclose(obs1.gti.met_stop.value, [4.0, 8.0, 15.0])


@requires_data("gammapy-data")
def test_datasets_stack_reduce():
    datasets = Datasets()
    obs_ids = [23523, 23526, 23559, 23592]

    for obs_id in obs_ids:
        filename = f"$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs{obs_id}.fits"
        ds = SpectrumDatasetOnOff.read(filename)
        datasets.append(ds)

    stacked = datasets.stack_reduce(name="stacked")

    assert_allclose(stacked.exposure.meta["livetime"].to_value("s"), 6313.8116406202325)

    info_table = datasets.info_table()
    assert_allclose(info_table["counts"], [124, 126, 119, 90])

    info_table_cum = datasets.info_table(cumulative=True)
    assert_allclose(info_table_cum["counts"], [124, 250, 369, 459])
    assert stacked.name == "stacked"


@requires_data("gammapy-data")
def test_datasets_stack_reduce_no_off():
    datasets = Datasets()
    obs_ids = [23523, 23526, 23559, 23592]

    for obs_id in obs_ids:
        filename = f"$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs{obs_id}.fits"
        ds = SpectrumDatasetOnOff.read(filename)
        datasets.append(ds)

    datasets[-1].counts_off = None

    with pytest.raises(ValueError):
        stacked = datasets.stack_reduce(name="stacked")

    datasets[-1].mask_safe.data[...] = False
    stacked = datasets.stack_reduce(name="stacked")
    assert_allclose(stacked.exposure.meta["livetime"].to_value("s"), 4732.5469999)
    assert stacked.counts == 369

    datasets[0].mask_safe.data[...] = False

    stacked = datasets.stack_reduce(name="stacked")
    assert_allclose(stacked.exposure.meta["livetime"].to_value("s"), 3150.81024152)
    assert stacked.counts == 245


@requires_data("gammapy-data")
def test_stack_livetime():
    dataset_ref = SpectrumDatasetOnOff.read(
        "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits"
    )

    energy_axis = dataset_ref.counts.geom.axes["energy"]
    energy_axis_true = dataset_ref.exposure.geom.axes["energy_true"]

    geom = RegionGeom(region=None, axes=[energy_axis])

    dataset = SpectrumDatasetOnOff.create(geom=geom, energy_axis_true=energy_axis_true)

    dataset.stack(dataset_ref)
    assert_allclose(dataset.exposure.meta["livetime"], 1581.736758 * u.s)

    dataset.stack(dataset_ref)
    assert_allclose(dataset.exposure.meta["livetime"], 2 * 1581.736758 * u.s)


def test_spectrum_dataset_on_off_to_yaml(tmpdir):
    spectrum_datasets_on_off = make_observation_list()
    datasets = Datasets(spectrum_datasets_on_off)
    datasets.write(
        filename=tmpdir / "datasets.yaml", filename_models=tmpdir / "models.yaml"
    )

    datasets_read = Datasets.read(
        filename=tmpdir / "datasets.yaml", filename_models=tmpdir / "models.yaml"
    )

    assert len(datasets_read) == len(datasets)
    assert datasets_read[0].name == datasets[0].name
    assert datasets_read[1].name == datasets[1].name
    assert datasets_read[1].counts.data.sum() == datasets[1].counts.data.sum()


class TestFit:
    """Test fit on counts spectra without any IRFs"""

    def setup_method(self):
        self.nbins = 30
        energy = np.logspace(-1, 1, self.nbins + 1) * u.TeV
        self.source_model = SkyModel(
            spectral_model=PowerLawSpectralModel(
                index=2, amplitude=1e5 * u.Unit("cm-2 s-1 TeV-1"), reference=0.1 * u.TeV
            )
        )
        bkg_model = PowerLawSpectralModel(
            index=3, amplitude=1e4 * u.Unit("cm-2 s-1 TeV-1"), reference=0.1 * u.TeV
        )

        self.alpha = 0.1
        random_state = get_random_state(23)
        npred = self.source_model.spectral_model.integral(energy[:-1], energy[1:]).value
        source_counts = random_state.poisson(npred)

        axis = MapAxis.from_edges(energy, name="energy", interp="log")
        geom = RegionGeom(region=None, axes=[axis])

        self.src = RegionNDMap.from_geom(geom=geom, data=source_counts)
        self.exposure = RegionNDMap.from_geom(geom.as_energy_true, data=1, unit="cm2 s")

        npred_bkg = bkg_model.integral(energy[:-1], energy[1:]).value

        bkg_counts = random_state.poisson(npred_bkg)
        off_counts = random_state.poisson(npred_bkg * 1.0 / self.alpha)
        self.bkg = RegionNDMap.from_geom(geom=geom, data=bkg_counts)
        self.off = RegionNDMap.from_geom(geom=geom, data=off_counts)

    def test_cash(self):
        """Simple CASH fit to the on vector"""
        dataset = SpectrumDataset(
            models=self.source_model,
            counts=self.src,
            exposure=self.exposure,
        )

        npred = dataset.npred().data
        assert_allclose(npred[5], 660.5171, rtol=1e-5)

        stat_val = dataset.stat_sum()
        assert_allclose(stat_val, -107346.5291, rtol=1e-5)

        self.source_model.parameters["index"].value = 1.12

        fit = Fit()
        fit.run(datasets=[dataset])

        # These values are check with sherpa fits, do not change
        pars = self.source_model.parameters
        assert_allclose(pars["index"].value, 1.995525, rtol=1e-3)
        assert_allclose(pars["amplitude"].value, 100245.9, rtol=1e-3)

    def test_wstat(self):
        """WStat with on source and background spectrum"""
        on_vector = self.src.copy()
        on_vector.data += self.bkg.data
        acceptance = RegionNDMap.from_geom(self.src.geom, data=1)
        acceptance_off = RegionNDMap.from_geom(self.bkg.geom, data=1 / self.alpha)

        dataset = SpectrumDatasetOnOff(
            counts=on_vector,
            counts_off=self.off,
            exposure=self.exposure,
            acceptance=acceptance,
            acceptance_off=acceptance_off,
        )
        dataset.models = self.source_model

        self.source_model.parameters.index = 1.12

        fit = Fit()
        result = fit.run(datasets=[dataset])
        pars = self.source_model.parameters

        assert_allclose(pars["index"].value, 1.997342, rtol=1e-3)
        assert_allclose(pars["amplitude"].value, 100245.187067, rtol=1e-3)
        assert_allclose(result.total_stat, 30.022316, rtol=1e-3)

    def test_fit_range(self):
        """Test fit range without complication of thresholds"""
        geom = self.src.geom
        mask_safe = RegionNDMap.from_geom(geom, dtype=bool)
        mask_safe.data += True

        dataset = SpectrumDatasetOnOff(counts=self.src, mask_safe=mask_safe)

        assert np.sum(dataset.mask_safe) == self.nbins
        energy_min, energy_max = dataset.energy_range

        assert_allclose(energy_max, 10)
        assert_allclose(energy_min, 0.1)

    def test_stat_profile(self):
        geom = self.src.geom
        mask_safe = RegionNDMap.from_geom(geom, dtype=bool)
        mask_safe.data += True

        dataset = SpectrumDataset(
            models=self.source_model,
            exposure=self.exposure,
            counts=self.src,
            mask_safe=mask_safe,
        )
        fit = Fit()
        fit.run(datasets=[dataset])
        true_idx = self.source_model.parameters["index"].value

        values = np.linspace(0.95 * true_idx, 1.05 * true_idx, 100)
        self.source_model.spectral_model.index.scan_values = values

        profile = fit.stat_profile(datasets=[dataset], parameter="index")
        actual = values[np.argmin(profile["stat_scan"])]
        assert_allclose(actual, true_idx, rtol=0.01)


def test_stat_sum():
    axis = MapAxis.from_energy_bounds(0.1, 10, 5, unit="TeV")
    geom = RegionGeom.create(None, axes=[axis])
    dataset = SpectrumDatasetOnOff.create(geom)
    dataset.counts_off = None

    stat = dataset.stat_sum()
    assert stat == 0

    dataset.mask_safe.data[0] = True
    with pytest.raises(AttributeError):
        dataset.stat_sum()


@requires_data("gammapy-data")
def test_priors():
    datasets = Datasets()
    obs_ids = [23523, 23526]

    for obs_id in obs_ids:
        filename = f"$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs{obs_id}.fits"
        ds = SpectrumDatasetOnOff.read(filename)
        datasets.append(ds)

    spectral_model = PowerLawSpectralModel(
        index=2.5, amplitude="3.8e-11 cm-2 s-1 TeV-1"
    )
    source_model = SkyModel(spectral_model=spectral_model, name="source")

    datasets.models = [source_model]

    stat_sum = datasets.stat_sum()

    spectral_model.index.prior = GaussianPrior(
        mu=spectral_model.index.value + 0.1, sigma=0.1
    )

    prior_stat_sum = spectral_model.index.prior_stat_sum()

    stat_sum_with_priors = datasets.stat_sum()

    assert_allclose(stat_sum, 87.928542, atol=1e-1)
    # Here we check that the prior is applied only once
    assert_allclose(stat_sum_with_priors, stat_sum + prior_stat_sum)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utilities to create scripts and command-line tools."""

import ast
import codecs
import operator
import os.path
import functools
import types
import warnings
import numpy as np
from base64 import urlsafe_b64encode
from pathlib import Path
from uuid import uuid4
import yaml
from gammapy.utils.check import add_checksum, verify_checksum

__all__ = [
    "from_yaml",
    "get_images_paths",
    "make_path",
    "read_yaml",
    "recursive_merge_dicts",
    "to_yaml",
    "write_yaml",
]

PATH_DOCS = Path(__file__).resolve().parent / ".." / ".." / "docs"
SKIP = ["_static", "_build", "_checkpoints", "docs/user-guide/model-gallery/"]
YAML_FORMAT = dict(sort_keys=False, indent=4, width=80, default_flow_style=False)


def get_images_paths(folder=PATH_DOCS):
    """Generator yields a Path for each image used in notebook.

    Parameters
    ----------
    folder : str
        Folder where to search.
    """
    for i in Path(folder).rglob("images/*"):
        if not any(s in str(i) for s in SKIP):
            yield i.resolve()


def from_yaml(text, sort_keys=False, checksum=False):
    """Read YAML file.

    Parameters
    ----------
    text : str
        yaml str
    sort_keys : bool, optional
        Whether to sort keys. Default is False.
    checksum : bool
        Whether to perform checksum verification. Default is False.

    Returns
    -------
    data : dict
        YAML file content as a dictionary.

    """
    data = yaml.safe_load(text)
    checksum_str = data.pop("checksum", None)
    if checksum:
        yaml_format = YAML_FORMAT.copy()
        yaml_format["sort_keys"] = sort_keys
        yaml_str = yaml.dump(data, **yaml_format)
        if not verify_checksum(yaml_str, checksum_str):
            warnings.warn("Checksum verification failed.", UserWarning)

    return data


def read_yaml(filename, logger=None, checksum=False):
    """Read YAML file.

    Parameters
    ----------
    filename : `~pathlib.Path`
        Filename.
    logger : `~logging.Logger`
        Logger.
    checksum : bool
        Whether to perform checksum verification. Default is False.

    Returns
    -------
    data : dict
        YAML file content as a dictionary.
    """
    path = make_path(filename)
    if logger is not None:
        logger.info(f"Reading {path}")

    text = path.read_text()
    return from_yaml(text, checksum=checksum)


def to_yaml(dictionary, sort_keys=False):
    """Dictionary to yaml file.

    Parameters
    ----------
    dictionary : dict
        Python dictionary.
    sort_keys : bool, optional
        Whether to sort keys. Default is False.
    """
    from gammapy.utils.metadata import CreatorMetaData

    yaml_format = YAML_FORMAT.copy()
    yaml_format["sort_keys"] = sort_keys
    text = yaml.safe_dump(dictionary, **yaml_format)
    creation = CreatorMetaData()
    return text + creation.to_yaml()


def write_yaml(
    text, filename, logger=None, sort_keys=False, checksum=False, overwrite=False
):
    """Write YAML file.

    Parameters
    ----------
    text : str
        yaml str
    filename : `~pathlib.Path`
        Filename.
    logger : `~logging.Logger`, optional
        Logger. Default is None.
    sort_keys : bool, optional
        Whether to sort keys. Default is True.
    checksum : bool, optional
        Whether to add checksum keyword. Default is False.
    overwrite : bool, optional
        Overwrite existing file. Default is False.
    """
    if checksum:
        text = add_checksum(text, sort_keys=sort_keys)

    path = make_path(filename)
    path.parent.mkdir(exist_ok=True)
    if path.exists() and not overwrite:
        raise IOError(f"File exists already: {path}")
    if logger is not None:
        logger.info(f"Writing {path}")
    path.write_text(text)


def make_name(name=None):
    """Make a dataset name."""
    if name is None:
        name = urlsafe_b64encode(codecs.decode(uuid4().hex, "hex")).decode()[:8]
        while name[0] == "_":
            name = urlsafe_b64encode(codecs.decode(uuid4().hex, "hex")).decode()[:8]

    if not isinstance(name, str):
        raise ValueError(
            "Name argument must be a string, "
            f"got '{name}', which is of type '{type(name)}'"
        )

    return name


def make_path(path):
    """Expand environment variables on `~pathlib.Path` construction.

    Parameters
    ----------
    path : str, `pathlib.Path`
        Path to expand.
    """
    # TODO: raise error or warning if environment variables that don't resolve are used
    # e.g. "spam/$DAMN/ham" where `$DAMN` is not defined
    # Otherwise this can result in cryptic errors later on
    if path is None:
        return None
    else:
        return Path(os.path.expandvars(path))


def recursive_merge_dicts(a, b):
    """Recursively merge two dictionaries.

    Entries in 'b' override entries in 'a'. The built-in update function cannot be
    used for hierarchical dicts, see:
    http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth/3233356#3233356

    Parameters
    ----------
    a : dict
        Dictionary to be merged.
    b : dict
        Dictionary to be merged.

    Returns
    -------
    c : dict
        Merged dictionary.

    Examples
    --------
    >>> from gammapy.utils.scripts import recursive_merge_dicts
    >>> a = dict(a=42, b=dict(c=43, e=44))
    >>> b = dict(d=99, b=dict(c=50, g=98))
    >>> c = recursive_merge_dicts(a, b)
    >>> print(c)
    {'a': 42, 'b': {'c': 50, 'e': 44, 'g': 98}, 'd': 99}
    """
    c = a.copy()
    for k, v in b.items():
        if k in c and isinstance(c[k], dict):
            c[k] = recursive_merge_dicts(c[k], v)
        else:
            c[k] = v
    return c


def requires_module(module_name):
    """
    Decorator that conditionally enables a method or property based on the availability of a module.

    If the specified module is available, the decorated method or property is returned as-is.
    If the module is not available:
      - For methods: replaces the method with one that raises ImportError when called.
      - For properties: replaces the property with one that raises ImportError when accessed.

    Parameters
    ----------
    module_name : str
        The name of the module to check for.

    Returns
    -------
    function or property
        The original object if the module is available, otherwise a fallback.
    """

    def decorator(obj):
        try:
            __import__(module_name)
            return obj  # Module is available
        except ImportError:
            if isinstance(obj, property):
                return property(
                    lambda self: raise_import_error(module_name, is_property=True)
                )
            elif isinstance(obj, (types.FunctionType, types.MethodType)):

                @functools.wraps(obj)
                def wrapper(*args, **kwargs):
                    raise_import_error(module_name)

                return wrapper
            else:
                raise TypeError(
                    "requires_module can only be used on methods or properties."
                )

    return decorator


def raise_import_error(module_name, is_property=False):
    """
    Raises an ImportError with a descriptive message about a missing module.

    Parameters
    ----------
    module_name : str
        The name of the required module.
    is_property : bool
        Whether the error is for a property (affects the error message).
    """
    kind = "property" if is_property else "method"
    raise ImportError(f"The '{module_name}' module is required to use this {kind}.")


# Mapping of AST operators to NumPy functions
_OPERATORS = {
    ast.And: np.logical_and,
    ast.Or: np.logical_or,
    ast.Eq: operator.eq,
    ast.NotEq: operator.ne,
    ast.Lt: operator.lt,
    ast.LtE: operator.le,
    ast.Gt: operator.gt,
    ast.GtE: operator.ge,
    ast.In: lambda a, b: np.isin(a, b),
    ast.NotIn: lambda a, b: ~np.isin(a, b),
}


def logic_parser(table, expression):
    """
    Parse and apply a logical expression to filter rows from an Astropy Table.

    This function evaluates a logical expression on each row of the input table
    and returns a new table containing only the rows that satisfy the expression.
    The expression can reference any column in the table by name and supports
    logical operators (`and`, `or`), comparison operators (`<`, `>`, `==`, `!=`, `in`),
    lists, and constants.

    Parameters
    ----------
    table : `~astropy.table.Table`
        The input table to filter.
    expression : str
        The logical expression to evaluate on each row. The expression can reference
        any column in the table by name.

    Returns
    -------
    table : `~astropy.table.Table`
        A table view containing only the rows that satisfy the expression. If no rows
        match the condition, an empty table with the same column names and data types
        as the input table is returned.

    Examples
    --------
    Given a table with columns 'OBS_ID' and 'EVENT_TYPE':

    >>> from astropy.table import Table
    >>> data = {'OBS_ID': [1, 2, 3, 4], 'EVENT_TYPE': ['1', '3', '4', '2']}
    >>> table = Table(data)
    >>> expression = '(OBS_ID < 3) and (OBS_ID > 1) and ((EVENT_TYPE in ["3", "4"]) or (EVENT_TYPE == "3"))'
    >>> filtered_table = logic_parser(table, expression)
    >>> print(filtered_table)
    OBS_ID EVENT_TYPE
    ------ ----------
         2          3

    """

    def eval_node(node):
        if isinstance(node, ast.BoolOp):
            op_func = _OPERATORS[type(node.op)]
            values = [eval_node(v) for v in node.values]
            result = values[0]
            for v in values[1:]:
                result = op_func(result, v)
            return result
        elif isinstance(node, ast.Compare):
            left = eval_node(node.left)
            for op, comparator in zip(node.ops, node.comparators):
                right = eval_node(comparator)
                op_func = _OPERATORS[type(op)]
                left = op_func(left, right)
            return left
        elif isinstance(node, ast.Name):
            if node.id not in table.colnames:
                raise KeyError(
                    f"Column '{node.id}' not found in the table. Available columns: {table.colnames}"
                )
            return table[node.id]
        elif isinstance(node, ast.Constant):
            return node.value
        elif isinstance(node, ast.List):
            return [eval_node(elt) for elt in node.elts]
        else:
            raise ValueError(f"Unsupported expression type: {type(node)}")

    expr_ast = ast.parse(expression, mode="eval")
    mask = eval_node(expr_ast.body)
    return table[mask]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.nddata import NoOverlapError
from astropy.time import Time
from regions import PointSkyRegion
from gammapy.maps import HpxNDMap, Map, MapAxis, RegionNDMap
from gammapy.maps.hpx.io import HPX_FITS_CONVENTIONS, HpxConv
from gammapy.utils.scripts import make_path, make_name
from gammapy.utils.time import time_ref_from_dict, time_ref_to_dict
from . import LightCurveTemplateTemporalModel, Models, SkyModel, TemplateSpatialModel

__all__ = ["read_hermes_cube"]


def _template_model_from_cta_sdc(filename, t_ref=None):
    """To create a `LightCurveTemplateTemporalModel` from the energy-dependent temporal model files of the cta-sdc1.

    This format is subject to change.
    """
    filename = str(make_path(filename))
    with fits.open(filename) as hdul:
        frame = hdul[0].header.get("frame", "icrs")
        position = SkyCoord(
            hdul[0].header["LONG"] * u.deg, hdul[0].header["LAT"] * u.deg, frame=frame
        )

        energy_hdu = hdul["ENERGIES"]
        energy_axis = MapAxis.from_nodes(
            nodes=energy_hdu.data,
            unit=energy_hdu.header["TUNIT1"],
            name="energy",
            interp="log",
        )
        time_hdu = hdul["TIMES"]
        time_header = time_hdu.header

        if t_ref is None:
            t_ref = Time(55555.5, format="mjd", scale="tt")
        time_header.update(time_ref_to_dict(t_ref, t_ref.scale))
        time_min = time_hdu.data["Initial Time"]
        time_max = time_hdu.data["Final Time"]
        edges = np.append(time_min, time_max[-1]) * u.Unit(time_header["TUNIT1"])
        data = hdul["SPECTRA"]

        time_ref = time_ref_from_dict(time_header, scale=t_ref.scale)
        time_axis = MapAxis.from_edges(edges=edges, name="time", interp="log")

        reg_map = RegionNDMap.create(
            region=PointSkyRegion(center=position),
            axes=[energy_axis, time_axis],
            data=np.array(list(data.data) * u.Unit(data.header["UNITS"])),
        )
    return LightCurveTemplateTemporalModel(reg_map, t_ref=time_ref, filename=filename)


def read_hermes_cube(filename):
    """Read 3d templates produced with hermes."""
    # add hermes conventions to the list used by gammapy
    hermes_conv = HpxConv(
        convname="hermes-template",
        colstring="TFLOAT",
        hduname="xtension",
        frame="COORDTYPE",
        quantity_type="differential",
    )
    HPX_FITS_CONVENTIONS["hermes-template"] = hermes_conv

    maps = []
    energy_nodes = []
    with fits.open(filename) as hdulist:
        # cannot read directly in 3d with Map.read because BANDS HDU is missing
        # https://gamma-astro-data-formats.readthedocs.io/en/v0.2/skymaps/index.html#bands-hdu
        # so we have to loop over hdus and create the energy axis
        for hdu in hdulist[1:]:
            template = HpxNDMap.from_hdu(hdu, format="hermes-template")
            # fix missing/incompatible infos
            template._unit = u.Unit(hdu.header["TUNIT1"])  # .from_hdu expect "BUNIT"
            if template.geom.frame == "G":
                template._geom._frame = "galactic"
            maps.append(template)
            energy_nodes.append(hdu.header["ENERGY"])  # SI unit (see header comment)
    # create energy axis and set unit
    energy_nodes *= u.Joule
    energy_nodes = energy_nodes.to("GeV")
    axis = MapAxis(
        energy_nodes, interp="log", name="energy_true", node_type="center", unit="GeV"
    )
    return Map.from_stack(maps, axis=axis)


def cutout_template_models(models, cutout_kwargs, datasets_names=None):
    """Apply cutout to template models.

    Parameters
    ----------
    models : `~gammapy.modeling.Models`
        List of models
    cutout_kwargs : dict
        Arguments passed to `gammap.map.cutout`
    datasets_names : list of str
        Names of the datasets to which the new model is applied.

    Returns
    -------
    models_cut : `~gammapy.modeling.Models`
        Models with cutout
    """
    models_cut = Models()
    if models is None:
        return models_cut
    for m in models:
        if isinstance(m.spatial_model, TemplateSpatialModel):
            try:
                map_ = m.spatial_model.map.cutout(**cutout_kwargs)
            except (NoOverlapError, ValueError):
                continue
            template_cut = TemplateSpatialModel(
                map_,
                normalize=m.spatial_model.normalize,
            )
            model_cut = SkyModel(
                spatial_model=template_cut,
                spectral_model=m.spectral_model,
                datasets_names=datasets_names,
                name=m.name if not datasets_names else f"{m.name}_{make_name()}",
            )
            models_cut.append(model_cut)
        else:
            models_cut.append(m)
    return models_cut



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.table import Table
from astropy.time import Time
from gammapy.data import EventList
from gammapy.maps import HpxGeom, Map, MapAxis, WcsNDMap
from gammapy.utils.testing import requires_dependency


@pytest.fixture()
def events():
    t = Table()
    t["EVENT_ID"] = np.array([1, 5], dtype=np.uint16)
    t["RA"] = [5, 11] * u.deg
    t["DEC"] = [0, 0] * u.deg
    t["ENERGY"] = [10, 12] * u.TeV
    t["TIME"] = Time("2025-01-01") + [3, 4] * u.s
    return EventList(t)


def test_map_fill_events_wcs(events):
    # 2D map
    m = Map.create(npix=(2, 1), binsz=10)
    m.fill_events(events)
    assert_allclose(m.data, [[1, 0]])

    # 3D with energy axis
    axis = MapAxis.from_edges([9, 11, 13], name="energy", unit="TeV")
    m = Map.create(npix=(2, 1), binsz=10, axes=[axis])
    m.fill_events(events)
    assert m.data.sum() == 1
    assert_allclose(m.data[0, 0, 0], 1)

    weights = np.array([0.5, 1])
    m = Map.create(npix=(2, 1), binsz=10, axes=[axis])
    m.fill_events(events, weights=weights)
    assert_allclose(m.data.sum(), 0.5)


@requires_dependency("healpy")
def test_map_fill_events_hpx(events):
    # 2D map
    m = Map.from_geom(HpxGeom(1))
    m.fill_events(events)
    assert m.data[4] == 2

    # 3D with energy axis
    axis = MapAxis.from_edges([9, 11, 13], name="energy", unit="TeV")
    m = Map.from_geom(HpxGeom(1, axes=[axis]))
    m.fill_events(events)
    assert m.data[0, 4] == 1
    assert m.data[1, 4] == 1

    weights = np.array([0.5, 1])
    m = Map.from_geom(HpxGeom(1, axes=[axis]))
    m.fill_events(events, weights=weights)
    assert_allclose(m.data.sum(), 1.5)


def test_map_fill_events_keyerror(events):
    axis = MapAxis([0, 1, 2], name="nokey")
    m = WcsNDMap.create(binsz=0.1, npix=10, axes=[axis])
    with pytest.raises(KeyError):
        m.fill_events(events)


def test_map_fill_events_uint_column(events):
    # Check that unsigned int column works.
    # Regression test for https://github.com/gammapy/gammapy/issues/1620
    axis = MapAxis.from_edges([0, 3, 6], name="event_id")
    m = Map.create(npix=(2, 1), binsz=10, axes=[axis])
    m.fill_events(events)
    assert m.data.sum() == 1
    assert_allclose(m.data[0, 0, 0], 1)



"""
Light curves
============

Compute per-observation and nightly fluxes of four Crab nebula observations.

Prerequisites
-------------

-  Knowledge of the high level interface to perform data reduction, see the
   :doc:`/tutorials/starting/analysis_1` tutorial.

Context
-------

This tutorial presents how light curve extraction is performed in
gammapy, i.e. how to measure the flux of a source in different time
bins.

Cherenkov telescopes usually work with observing runs and distribute
data according to this basic time interval. A typical use case is to
look for variability of a source on various time bins: observation
run-wise binning, nightly, weekly etc.

**Objective: The Crab nebula is not known to be variable at TeV
energies, so we expect constant brightness within statistical and
systematic errors. Compute per-observation and nightly fluxes of the
four Crab nebula observations from the H.E.S.S. first public test data
release.**

Proposed approach
-----------------

We will demonstrate how to compute a light curve from 3D reduced
datasets (`~gammapy.datasets.MapDataset`) as well as 1D ON-OFF
spectral datasets (`~gammapy.datasets.SpectrumDatasetOnOff`).

The data reduction will be performed with the high level interface for
the data reduction. Then we will use the
`~gammapy.estimators.LightCurveEstimator` class, which is able to
extract a light curve independently of the dataset type.

"""

######################################################################
# Setup
# -----
#
# As usual, we’ll start with some general imports…
#

import logging
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.time import Time

# %matplotlib inline
import matplotlib.pyplot as plt
from IPython.display import display
from gammapy.analysis import Analysis, AnalysisConfig
from gammapy.estimators import LightCurveEstimator
from gammapy.modeling.models import (
    Models,
    PointSpatialModel,
    PowerLawSpectralModel,
    SkyModel,
)

log = logging.getLogger(__name__)


######################################################################
# Analysis configuration
# ----------------------
#
# For the 1D and 3D extraction, we will use the same CrabNebula
# configuration than in the :doc:`/tutorials/starting/analysis_1` tutorial
# using the high level interface of Gammapy.
#
# From the high level interface, the data reduction for those observations
# is performed as follows.
#


######################################################################
# Building the 3D analysis configuration
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#

conf_3d = AnalysisConfig()


######################################################################
# Definition of the data selection
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Here we use the Crab runs from the
# `H.E.S.S. DL3 data release 1 <https://www.mpi-hd.mpg.de/hfm/HESS/pages/dl3-dr1/>`__.
#

conf_3d.observations.obs_ids = [23523, 23526, 23559, 23592]


######################################################################
# Definition of the dataset geometry
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#

# We want a 3D analysis
conf_3d.datasets.type = "3d"

# We want to extract the data by observation and therefore to not stack them
conf_3d.datasets.stack = False

# Here is the WCS geometry of the Maps
conf_3d.datasets.geom.wcs.skydir = dict(
    frame="icrs", lon=83.63308 * u.deg, lat=22.01450 * u.deg
)
conf_3d.datasets.geom.wcs.binsize = 0.02 * u.deg
conf_3d.datasets.geom.wcs.width = dict(width=1 * u.deg, height=1 * u.deg)

# We define a value for the IRF Maps binsize
conf_3d.datasets.geom.wcs.binsize_irf = 0.2 * u.deg

# Define energy binning for the Maps
conf_3d.datasets.geom.axes.energy = dict(min=0.7 * u.TeV, max=10 * u.TeV, nbins=5)
conf_3d.datasets.geom.axes.energy_true = dict(min=0.3 * u.TeV, max=20 * u.TeV, nbins=20)


######################################################################
# Run the 3D data reduction
# ~~~~~~~~~~~~~~~~~~~~~~~~~
#

analysis_3d = Analysis(conf_3d)
analysis_3d.get_observations()
analysis_3d.get_datasets()


######################################################################
# Define the model to be used
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Here we don’t try to fit the model parameters to the whole dataset, but
# we use predefined values instead.
#

target_position = SkyCoord(ra=83.63308, dec=22.01450, unit="deg")
spatial_model = PointSpatialModel(
    lon_0=target_position.ra, lat_0=target_position.dec, frame="icrs"
)

spectral_model = PowerLawSpectralModel(
    index=2.702,
    amplitude=4.712e-11 * u.Unit("1 / (cm2 s TeV)"),
    reference=1 * u.TeV,
)

sky_model = SkyModel(
    spatial_model=spatial_model, spectral_model=spectral_model, name="crab"
)


######################################################################
# We assign them the model to be fitted to each dataset
#

models = Models([sky_model])
analysis_3d.set_models(models)


######################################################################
# Light Curve estimation by observation
# -------------------------------------
#
# We can now create the light curve estimator.
#
# We pass it the list of datasets and the name of the model component for
# which we want to build the light curve. In a given time bin, the only
# free parameter of the source is its normalization. We can optionally ask
# for parameters of other model components to be reoptimized during fit,
# that is most of the time to fit background normalization in each time
# bin.
#
# If we don’t set any time interval, the
# `~gammapy.estimators.LightCurveEstimator` determines the flux of
# each dataset and places it at the corresponding time in the light curve.
# Here one dataset equals to one observing run.
#

lc_maker_3d = LightCurveEstimator(
    energy_edges=[1, 10] * u.TeV, source="crab", reoptimize=False
)
# Example showing how to change some parameters from the object itself
lc_maker_3d.n_sigma_ul = 3  # Number of sigma to use for upper limit computation
lc_maker_3d.selection_optional = (
    "all"  # Add the computation of upper limits and likelihood profile
)
lc_3d = lc_maker_3d.run(analysis_3d.datasets)


######################################################################
# The lightcurve `~gammapy.estimators.FluxPoints` object ``lc_3d`` contains
# a table which we can explore.
#
# Example showing how to change just before plotting the threshold on the signal significance
# (points vs upper limits), even if this has no effect with this data set.

fig, ax = plt.subplots(
    figsize=(8, 6),
    gridspec_kw={"left": 0.16, "bottom": 0.2, "top": 0.98, "right": 0.98},
)
lc_3d.sqrt_ts_threshold_ul = 5
lc_3d.plot(ax=ax, axis_name="time")
plt.show()

table = lc_3d.to_table(format="lightcurve", sed_type="flux")
display(table["time_min", "time_max", "e_min", "e_max", "flux", "flux_err"])


######################################################################
# Running the light curve extraction in 1D
# ----------------------------------------
#


######################################################################
# Building the 1D analysis configuration
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#

conf_1d = AnalysisConfig()


######################################################################
# Definition of the data selection
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Here we use the Crab runs from the
# `H.E.S.S. DL3 data release 1 <https://www.mpi-hd.mpg.de/hfm/HESS/pages/dl3-dr1/>`__
#

conf_1d.observations.obs_ids = [23523, 23526, 23559, 23592]


######################################################################
# Definition of the dataset geometry
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#

# We want a 1D analysis
conf_1d.datasets.type = "1d"

# We want to extract the data by observation and therefore to not stack them
conf_1d.datasets.stack = False

# Here we define the ON region and make sure that PSF leakage is corrected
conf_1d.datasets.on_region = dict(
    frame="icrs",
    lon=83.63308 * u.deg,
    lat=22.01450 * u.deg,
    radius=0.1 * u.deg,
)
conf_1d.datasets.containment_correction = True

# Finally we define the energy binning for the spectra
conf_1d.datasets.geom.axes.energy = dict(min=0.7 * u.TeV, max=10 * u.TeV, nbins=5)
conf_1d.datasets.geom.axes.energy_true = dict(min=0.3 * u.TeV, max=20 * u.TeV, nbins=20)


######################################################################
# Run the 1D data reduction
# ~~~~~~~~~~~~~~~~~~~~~~~~~
#

analysis_1d = Analysis(conf_1d)
analysis_1d.get_observations()
analysis_1d.get_datasets()


######################################################################
# Define the model to be used
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Here we don’t try to fit the model parameters to the whole dataset, but
# we use predefined values instead.
#

target_position = SkyCoord(ra=83.63308, dec=22.01450, unit="deg")

spectral_model = PowerLawSpectralModel(
    index=2.702,
    amplitude=4.712e-11 * u.Unit("1 / (cm2 s TeV)"),
    reference=1 * u.TeV,
)

sky_model = SkyModel(spectral_model=spectral_model, name="crab")


######################################################################
# We assign the model to be fitted to each dataset. We can use the same
# `~gammapy.modeling.models.SkyModel` as before.
#

models = Models([sky_model])
analysis_1d.set_models(models)


######################################################################
# Extracting the light curve
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
#

lc_maker_1d = LightCurveEstimator(
    energy_edges=[1, 10] * u.TeV, source="crab", reoptimize=False
)
lc_1d = lc_maker_1d.run(analysis_1d.datasets)

print(lc_1d.geom.axes.names)

display(lc_1d.to_table(sed_type="flux", format="lightcurve"))


######################################################################
# Compare results
# ~~~~~~~~~~~~~~~
#
# Finally we compare the result for the 1D and 3D lightcurve in a single
# figure:
#

fig, ax = plt.subplots(
    figsize=(8, 6),
    gridspec_kw={"left": 0.16, "bottom": 0.2, "top": 0.98, "right": 0.98},
)
lc_1d.plot(ax=ax, marker="o", label="1D")
lc_3d.plot(ax=ax, marker="o", label="3D")
plt.legend()
plt.show()


######################################################################
# Night-wise LC estimation
# ------------------------
#
# Here we want to extract a night curve per night. We define the time
# intervals that cover the three nights.
#

time_intervals = [
    Time([53343.5, 53344.5], format="mjd", scale="utc"),
    Time([53345.5, 53346.5], format="mjd", scale="utc"),
    Time([53347.5, 53348.5], format="mjd", scale="utc"),
]


######################################################################
# To compute the LC on the time intervals defined above, we pass the
# `~gammapy.estimators.LightCurveEstimator` the list of time intervals.
#
# Internally, datasets are grouped per time interval and a flux extraction
# is performed for each group.
#

lc_maker_1d = LightCurveEstimator(
    energy_edges=[1, 10] * u.TeV,
    time_intervals=time_intervals,
    source="crab",
    reoptimize=False,
    selection_optional="all",
)

nightwise_lc = lc_maker_1d.run(analysis_1d.datasets)

fig, ax = plt.subplots(
    figsize=(8, 6),
    gridspec_kw={"left": 0.16, "bottom": 0.2, "top": 0.98, "right": 0.98},
)
nightwise_lc.plot(ax=ax, color="tab:orange")
nightwise_lc.plot_ts_profiles(ax=ax)
ax.set_ylim(1e-12, 3e-12)

plt.show()

######################################################################
# What next?
# ----------
#
# When sources are bright enough to look for variability at small time
# scales, the per-observation time binning is no longer relevant. One can
# easily extend the light curve estimation approach presented above to any
# time binning. This is demonstrated in the :doc:`/tutorials/analysis-time/light_curve_flare`
# tutorial. which shows the extraction of the lightcurve of an AGN flare.
#
